from __future__ import annotations

import sys
import typing as t

if t.TYPE_CHECKING:
    from .internal.types import ClientT


def loader(callback: t.Callable[[ClientT], None]) -> t.Callable[[ClientT], None]:
    """First-order decorator to set the load callback for this module.

    Usage
    -----
    ```py
    client.load_extension("my_extension")

    # In my_extension.py:

    @arc.loader
    def load(client: arc.Client) -> None:
        client.add_plugin(...)
    ```

    See Also
    --------
    - [`Client.load_extension`][arc.client.Client.load_extension]
    """
    module = sys.modules[callback.__module__]
    setattr(module, "__arc_extension_loader__", callback)
    return callback


def unloader(callback: t.Callable[[ClientT], None]) -> t.Callable[[ClientT], None]:
    """First-order decorator to set the unload callback for this module.

    Usage
    -----
    ```py
    client.unload_extension("my_extension")

    # In my_extension.py:

    @arc.unloader
    def unload(client: arc.Client) -> None:
        client.remove_plugin(...)
    ```

    See Also
    --------
    - [`Client.unload_extension`][arc.client.Client.unload_extension]
    """
    module = sys.modules[callback.__module__]
    setattr(module, "__arc_extension_unloader__", callback)
    return callback



from __future__ import annotations

import typing as t

import attr
import hikari

from ..context import Context
from ..errors import CommandInvokeError
from ..internal.types import ClientT, MessageContextCallbackT, ResponseBuilderT
from .base import AutodeferMode, CallableCommandBase

if t.TYPE_CHECKING:
    import asyncio

    from .base import CallableCommandProto

__all__ = ("MessageCommand", "message_command")


@attr.define(slots=True, kw_only=True)
class MessageCommand(CallableCommandBase[ClientT, hikari.api.ContextMenuCommandBuilder]):
    """A context menu command that is invoked by right-clicking a message."""

    @property
    def command_type(self) -> hikari.CommandType:
        return hikari.CommandType.MESSAGE

    @property
    def qualified_name(self) -> t.Sequence[str]:
        return (self.name,)

    def _get_context(
        self, interaction: hikari.CommandInteraction, command: CallableCommandProto[ClientT]
    ) -> Context[ClientT]:
        assert self.client is not None

        return Context(self.client, command, interaction)

    def _build(self) -> hikari.api.ContextMenuCommandBuilder:
        return hikari.impl.ContextMenuCommandBuilder(
            name=self.name,
            type=self.command_type,
            default_member_permissions=self.default_permissions,
            is_dm_enabled=self.is_dm_enabled,
            is_nsfw=self.is_nsfw,
            name_localizations=self.name_localizations,  # pyright: ignore reportGeneralTypeIssues
        )

    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> asyncio.Future[ResponseBuilderT] | None:
        if interaction.command_type is not hikari.CommandType.MESSAGE:
            raise CommandInvokeError("Cannot invoke a message command with a non-message command interaction.")

        assert interaction.resolved is not None and interaction.target_id is not None

        msg = interaction.resolved.messages[interaction.target_id]

        return await super().invoke(interaction, msg, *args, **kwargs)


def message_command(
    name: str,
    *,
    guilds: t.Sequence[hikari.SnowflakeishOr[hikari.PartialGuild]] | None = None,
    is_dm_enabled: bool = True,
    is_nsfw: bool = False,
    autodefer: bool | AutodeferMode = True,
    default_permissions: hikari.UndefinedOr[hikari.Permissions] = hikari.UNDEFINED,
    name_localizations: t.Mapping[hikari.Locale, str] | None = None,
) -> t.Callable[[MessageContextCallbackT[ClientT]], MessageCommand[ClientT]]:
    """A decorator that creates a context-menu command on a message.

    Parameters
    ----------
    name : str
        The name of the command.
    guilds : t.Sequence[hikari.SnowflakeishOr[hikari.PartialGuild]] | None
        The guilds this command is available in.
    is_dm_enabled : bool
        Whether this command is enabled in DMs.
    is_nsfw : bool
        Whether this command is NSFW.
    autodefer : bool | AutodeferMode
        If True, this command will be automatically deferred if it takes longer than 2 seconds to respond, by default True
    default_permissions : hikari.UndefinedOr[hikari.Permissions]
        The default permissions for this command.
        Keep in mind that guild administrators can change this, it should only be used to provide safe defaults.
    name_localizations : t.Mapping[hikari.Locale, str] | None
        The localizations for this command's name.

    Usage
    -----
    ```py
    @client.include
    @arc.message_command(name="Say Hi", description="Say hi!")
    async def hi_msg(
        ctx: arc.Context[arc.GatewayClient], message: hikari.Message
    ) -> None:
        await ctx.respond(f"Hey {message.author}!")
    ```
    """

    def decorator(callback: MessageContextCallbackT[ClientT]) -> MessageCommand[ClientT]:
        guild_ids = [hikari.Snowflake(guild) for guild in guilds] if guilds else []

        return MessageCommand(
            callback=callback,
            name=name,
            autodefer=AutodeferMode(autodefer),
            guilds=guild_ids,
            is_dm_enabled=is_dm_enabled,
            is_nsfw=is_nsfw,
            default_permissions=default_permissions,
            name_localizations=name_localizations or {},
        )

    return decorator


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import abc
import asyncio
import typing as t

import attr
import hikari

from ..abc import HasErrorHandler
from ..context import AutodeferMode
from ..internal.types import BuilderT, ClientT, CommandCallbackT, ResponseBuilderT

if t.TYPE_CHECKING:
    from ..context import Context
    from ..plugin import Plugin

__all__ = ("CommandProto", "CallableCommandProto", "CommandBase", "CallableCommandBase")


class CommandProto(t.Protocol):
    """A protocol for any command-like object. This includes commands, groups, subgroups, and subcommands."""

    name: str
    """The name of the command."""
    name_localizations: t.Mapping[hikari.Locale, str]
    """The name of the command in different locales."""

    @property
    @abc.abstractmethod
    def command_type(self) -> hikari.CommandType:
        """The type of command this object represents."""

    @property
    @abc.abstractmethod
    def qualified_name(self) -> t.Sequence[str]:
        """The fully qualified name of this command."""


class CallableCommandProto(t.Protocol, t.Generic[ClientT]):
    """A protocol for any command-like object that can be called directly. This includes commands and subcommands."""

    name: str
    """The name of the command."""
    name_localizations: t.Mapping[hikari.Locale, str]
    """The name of the command in different locales."""
    callback: CommandCallbackT[ClientT]

    @property
    @abc.abstractmethod
    def command_type(self) -> hikari.CommandType:
        """The type of command this object represents."""

    @property
    @abc.abstractmethod
    def qualified_name(self) -> t.Sequence[str]:
        """The fully qualified name of this command."""

    @abc.abstractmethod
    async def __call__(self, ctx: Context[ClientT], *args: t.Any, **kwargs: t.Any) -> None:
        """Invoke this command with the given context.

        Parameters
        ----------
        ctx : ContextT
            The context to invoke this command with.
        args : tuple[Any]
            The positional arguments to pass to the callback.
        kwargs : dict[str, Any]
            The keyword arguments to pass to the callback.
        """

    @abc.abstractmethod
    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> None | asyncio.Future[ResponseBuilderT]:
        """Invoke this command with the given context.

        Parameters
        ----------
        interaction : hikari.CommandInteraction
            The interaction to invoke this command with.
        args : tuple[Any]
            The positional arguments to pass to the callback.
        kwargs : dict[str, Any]
            The keyword arguments to pass to the callback.

        Returns
        -------
        None | asyncio.Future[ResponseBuilderT]
            If this is a REST client, returns the response builder.

        Raises
        ------
        RuntimeError
            If this command has not been added to a client.
        """

    @abc.abstractmethod
    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        ...


@attr.define(slots=True, kw_only=True)
class CommandBase(HasErrorHandler[ClientT], t.Generic[ClientT, BuilderT]):
    """An abstract base class for all application commands."""

    name: str
    """The name of this command."""

    _client: ClientT | None = attr.field(init=False, default=None)
    """The client that is handling this command."""

    _plugin: Plugin[ClientT] | None = attr.field(init=False, default=None)
    """The plugin that this command belongs to, if any."""

    guilds: hikari.UndefinedOr[t.Sequence[hikari.Snowflake]] = hikari.UNDEFINED
    """The guilds this command is available in."""

    autodefer: AutodeferMode = AutodeferMode.ON
    """If ON, this command will be automatically deferred if it takes longer than 2 seconds to respond."""

    is_dm_enabled: bool = True
    """Whether this command is enabled in DMs."""

    default_permissions: hikari.UndefinedOr[hikari.Permissions] = hikari.UNDEFINED
    """The default permissions for this command.
    Keep in mind that guild administrators can change this, it should only be used to provide safe defaults."""

    name_localizations: t.Mapping[hikari.Locale, str] = attr.field(factory=dict)
    """The localizations for this command's name."""

    is_nsfw: bool = False
    """Whether this command is NSFW. If true, the command will only be available in NSFW channels."""

    _instances: dict[hikari.Snowflake | None, hikari.PartialCommand] = attr.field(factory=dict)
    """A mapping of guild IDs to command instances. None corresponds to the global instance, if any."""

    @property
    @abc.abstractmethod
    def command_type(self) -> hikari.CommandType:
        """The type of command this object represents."""

    @property
    @abc.abstractmethod
    def qualified_name(self) -> t.Sequence[str]:
        """The fully qualified name of this command."""

    @property
    def client(self) -> ClientT:
        """The client that is handling this command."""
        if self._client is None:
            raise RuntimeError(
                f"Command '{self.qualified_name}' was not included in a client, '{type(self).__name__}.client' cannot be accessed until it is included in a client."
            )
        return self._client

    @property
    def plugin(self) -> Plugin[ClientT] | None:
        """The plugin that this command belongs to, if any."""
        return self._plugin

    def _register_instance(
        self, instance: hikari.PartialCommand, guild: hikari.SnowflakeishOr[hikari.PartialGuild] | None = None
    ) -> None:
        self._instances[hikari.Snowflake(guild) if guild else None] = instance

    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        try:
            if self.error_handler is not None:
                await self.error_handler(ctx, exc)
            else:
                raise exc
        except Exception as exc:
            if self.plugin:
                await self.plugin._handle_exception(ctx, exc)
            else:
                await self.client._on_error(ctx, exc)

    async def publish(self, guild: hikari.SnowflakeishOr[hikari.PartialGuild] | None = None) -> hikari.PartialCommand:
        """Publish this command to the given guild, or globally if no guild is provided.

        Parameters
        ----------
        guild : hikari.Snowflakeish | None
            The guild to publish this command to. If None, publish globally.

        Returns
        -------
        hikari.PartialCommand
            The published command.
        """
        if self.client.application is None:
            raise RuntimeError("Cannot publish command without a client.")

        kwargs = self._to_dict()

        if self.command_type is hikari.CommandType.SLASH:
            created = await self.client.app.rest.create_slash_command(self.client.application, **kwargs)
        else:
            created = await self.client.app.rest.create_context_menu_command(
                self.client.application, type=self.command_type, **kwargs
            )

        self._instances[hikari.Snowflake(guild) if guild else None] = created

        return created

    async def unpublish(self, guild: hikari.SnowflakeishOr[hikari.PartialGuild] | None = None) -> None:
        """Unpublish this command from the given guild, or globally if no guild is provided.

        Parameters
        ----------
        guild : hikari.Snowflakeish | None
            The guild to unpublish this command from. If None, unpublish globally.
        """
        if command := self._instances.pop(hikari.Snowflake(guild) if guild else None, None):
            await command.delete()

    @abc.abstractmethod
    def _get_context(
        self, interaction: hikari.CommandInteraction, command: CallableCommandProto[ClientT]
    ) -> Context[ClientT]:
        """Create a context object from an interaction."""

    @abc.abstractmethod
    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> None | asyncio.Future[ResponseBuilderT]:
        """Invoke this command with the given context.

        Parameters
        ----------
        interaction : hikari.CommandInteraction
            The interaction to invoke this command with.
        args : tuple[Any]
            The positional arguments to pass to the callback.
        kwargs : dict[str, Any]
            The keyword arguments to pass to the callback.

        Returns
        -------
        None | asyncio.Future[ResponseBuilderT]
            If this is a REST client, returns the response builder.

        Raises
        ------
        RuntimeError
            If this command has not been added to a client.
        """

    def _to_dict(self) -> dict[str, t.Any]:
        return {
            "name": self.name,
            "nsfw": self.is_nsfw,
            "default_member_permissions": self.default_permissions,
            "name_localizations": self.name_localizations,
            "dm_enabled": self.is_dm_enabled,
        }

    @abc.abstractmethod
    def _build(self) -> BuilderT:
        """Create a builder out of this command."""

    def _client_include_hook(self, client: ClientT) -> None:
        """Called when the client requests the command be added to it."""
        self._client = client
        self.client._add_command(self)

    def _plugin_include_hook(self, plugin: Plugin[ClientT]) -> None:
        """Called when the plugin requests the command be added to it."""
        self._plugin = plugin
        self._plugin._add_command(self)

    async def _handle_callback(
        self, command: CallableCommandProto[ClientT], ctx: Context[ClientT], *args: t.Any, **kwargs: t.Any
    ) -> None:
        try:
            await self.client.injector.call_with_async_di(command.callback, ctx, *args, **kwargs)
        except Exception as e:
            await command._handle_exception(ctx, e)

    # TODO - hooks, max_concurrency, cooldowns


@attr.define(slots=True, kw_only=True)
class CallableCommandBase(CommandBase[ClientT, BuilderT]):
    """A command that can be called directly. Note that this does not include subcommands, as those are options."""

    callback: CommandCallbackT[ClientT]
    """The callback to invoke when this command is called."""

    _invoke_task: asyncio.Task[t.Any] | None = attr.field(init=False, default=None)

    async def __call__(self, ctx: Context[ClientT], *args: t.Any, **kwargs: t.Any) -> None:
        await self.callback(ctx, *args, **kwargs)

    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> None | asyncio.Future[ResponseBuilderT]:
        ctx = self._get_context(interaction, self)
        if self.autodefer.should_autodefer:
            ctx._start_autodefer(self.autodefer)
        self._invoke_task = asyncio.create_task(self._handle_callback(self, ctx, *args, **kwargs))
        if self.client.is_rest:
            return ctx._resp_builder


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import abc
import functools
import importlib
import inspect
import logging
import pathlib
import sys
import traceback
import typing as t
from contextlib import suppress

import alluka
import hikari

from .command import MessageCommand, SlashCommand, SlashGroup, SlashSubCommand, SlashSubGroup, UserCommand
from .context import AutodeferMode, Context
from .errors import ExtensionLoadError, ExtensionUnloadError
from .events import CommandErrorEvent
from .internal.sync import _sync_commands
from .internal.types import AppT, BuilderT, EventCallbackT, EventT, ResponseBuilderT
from .plugin import Plugin

if t.TYPE_CHECKING:
    import typing_extensions as te

    from .command import CommandBase, SlashCommandLike

__all__ = ("Client", "GatewayClient", "RESTClient")


T = t.TypeVar("T")
P = t.ParamSpec("P")

logger = logging.getLogger(__name__)


class Client(t.Generic[AppT], abc.ABC):
    """A base class for an `arc` client.
    See [`GatewayClient`][arc.client.GatewayClient] and [`RESTClient`][arc.client.RESTClient] for implementations.

    Parameters
    ----------
    app : AppT
        The application this client is for.
    default_enabled_guilds : t.Sequence[hikari.Snowflake] | None, optional
        The guilds that slash commands will be registered in by default, by default None
    autosync : bool, optional
        Whether to automatically sync commands on startup, by default True
    """

    __slots__: t.Sequence[str] = (
        "_app",
        "_default_enabled_guilds",
        "_application",
        "_slash_commands",
        "_message_commands",
        "_user_commands",
        "_injector",
        "_autosync",
        "_plugins",
        "_loaded_extensions",
    )

    def __init__(
        self, app: AppT, *, default_enabled_guilds: t.Sequence[hikari.Snowflake] | None = None, autosync: bool = True
    ) -> None:
        self._app = app
        self._default_enabled_guilds = default_enabled_guilds
        self._application: hikari.Application | None = None
        self._slash_commands: dict[str, SlashCommandLike[te.Self]] = {}
        self._message_commands: dict[str, MessageCommand[te.Self]] = {}
        self._user_commands: dict[str, UserCommand[te.Self]] = {}
        self._injector: alluka.Client = alluka.Client()
        self._plugins: dict[str, Plugin[te.Self]] = {}
        self._loaded_extensions: list[str] = []
        self._autosync = autosync

    @property
    @abc.abstractmethod
    def is_rest(self) -> bool:
        """Whether the app is a rest client or a gateway client.

        This controls the client response flow, if True, `Client.handle_command_interaction` and `Client.handle_autocomplete_interaction`
        will return interaction response builders to be sent back to Discord, otherwise they will return None.
        """

    @property
    def app(self) -> AppT:
        """The application this client is for."""
        return self._app

    @property
    def rest(self) -> hikari.api.RESTClient:
        """The REST client of the underyling app."""
        return self.app.rest

    @property
    def application(self) -> hikari.Application | None:
        """The application object for this client. This is fetched on startup."""
        return self._application

    @property
    def injector(self) -> alluka.Client:
        """The injector for this client."""
        return self._injector

    @property
    def default_enabled_guilds(self) -> t.Sequence[hikari.Snowflake] | None:
        """The guilds that slash commands will be registered in by default."""
        return self._default_enabled_guilds

    @property
    def commands(self) -> t.Mapping[hikari.CommandType, t.Mapping[str, CommandBase[te.Self, t.Any]]]:
        """All commands added to this client, categorized by command type."""
        return {
            hikari.CommandType.SLASH: self.slash_commands,
            hikari.CommandType.MESSAGE: self._message_commands,
            hikari.CommandType.USER: self._user_commands,
        }

    @property
    def slash_commands(self) -> t.Mapping[str, SlashCommandLike[te.Self]]:
        """The slash commands added to this client. This only includes top-level commands and groups."""
        return self._slash_commands

    @property
    def message_commands(self) -> t.Mapping[str, MessageCommand[te.Self]]:
        """The message commands added to this client."""
        return self._message_commands

    @property
    def user_commands(self) -> t.Mapping[str, UserCommand[te.Self]]:
        """The user commands added to this client."""
        return self._user_commands

    @property
    def plugins(self) -> t.Mapping[str, Plugin[te.Self]]:
        """The plugins added to this client."""
        return self._plugins

    def _add_command(self, command: CommandBase[te.Self, t.Any]) -> None:
        """Add a command to this client. Called by include hooks."""
        if isinstance(command, (SlashCommand, SlashGroup)):
            self._add_slash_command(command)
        elif isinstance(command, MessageCommand):
            self._add_message_command(command)
        elif isinstance(command, UserCommand):
            self._add_user_command(command)

    def _remove_command(self, command: CommandBase[te.Self, t.Any]) -> None:
        """Remove a command from this client. Called by remove hooks."""
        if isinstance(command, (SlashCommand, SlashGroup)):
            self._slash_commands.pop(command.name, None)
        elif isinstance(command, MessageCommand):
            self._message_commands.pop(command.name, None)
        elif isinstance(command, UserCommand):
            self._user_commands.pop(command.name, None)

    def _add_slash_command(self, command: SlashCommandLike[te.Self]) -> None:
        """Add a slash command to this client."""
        if self.slash_commands.get(command.name) is not None:
            logger.warning(
                f"Shadowing already registered slash command '{command.name}'. Did you define multiple commands/groups with the same name?"
            )

        self._slash_commands[command.name] = command

    def _add_message_command(self, command: MessageCommand[te.Self]) -> None:
        """Add a message command to this client."""
        if self._message_commands.get(command.name) is not None:
            logger.warning(
                f"Shadowing already registered message command '{command.name}'. Did you define multiple commands with the same name?"
            )

        self._message_commands[command.name] = command

    def _add_user_command(self, command: UserCommand[te.Self]) -> None:
        """Add a user command to this client."""
        if self._user_commands.get(command.name) is not None:
            logger.warning(
                f"Shadowing already registered user command '{command.name}'. Did you define multiple commands with the same name?"
            )

        self._user_commands[command.name] = command

    async def _on_startup(self) -> None:
        """Called when the client is starting up.
        Fetches application, syncs commands, calls user-defined startup.
        """
        self._application = await self.app.rest.fetch_application()
        logger.debug(f"Fetched application: '{self.application}'")
        if self._autosync:
            await _sync_commands(self)
        await self.on_startup()

    async def on_startup(self) -> None:
        """Called when the client is starting up.
        Override for custom startup logic.
        """

    async def _on_shutdown(self) -> None:
        """Called when the client is shutting down.
        Reserved for internal shutdown logic.
        """
        await self.on_shutdown()

    async def on_shutdown(self) -> None:
        """Called when the client is shutting down.
        Override for custom shutdown logic.
        """

    async def _on_error(self, ctx: Context[te.Self], exception: Exception) -> None:
        await self.on_error(ctx, exception)

    async def on_error(self, context: Context[te.Self], exception: Exception) -> None:
        """Called when an error occurs in a command callback and all other error handlers have failed.

        Parameters
        ----------
        context : Context[te.Self]
            The context of the command.
        exception : Exception
            The exception that was raised.
        """
        print(f"Unhandled error in command '{context.command.name}' callback: {exception}", file=sys.stderr)
        traceback.print_exception(type(exception), exception, exception.__traceback__, file=sys.stderr)
        with suppress(Exception):
            await context.respond("❌ Something went wrong. Please contact the bot developer.")

    async def on_command_interaction(self, interaction: hikari.CommandInteraction) -> ResponseBuilderT | None:
        """Should be called when a command interaction is sent by Discord.

        Parameters
        ----------
        interaction : hikari.CommandInteraction
            The interaction that was created.

        Returns
        -------
        ResponseBuilderT | None
            The response builder to send back to Discord, if using a REST client.
        """
        command = None

        if interaction.command_type is hikari.CommandType.SLASH:
            command = self.slash_commands.get(interaction.command_name)
        elif interaction.command_type is hikari.CommandType.MESSAGE:
            command = self.message_commands.get(interaction.command_name)
        elif interaction.command_type is hikari.CommandType.USER:
            command = self.user_commands.get(interaction.command_name)

        if command is None:
            logger.warning(f"Received interaction for unknown command '{interaction.command_name}'.")
            return

        fut = await command.invoke(interaction)

        if fut is not None:
            return await fut

    async def on_autocomplete_interaction(
        self, interaction: hikari.AutocompleteInteraction
    ) -> hikari.api.InteractionAutocompleteBuilder | None:
        """Should be called when an autocomplete interaction is sent by Discord.

        Parameters
        ----------
        interaction : hikari.AutocompleteInteraction
            The interaction that was created.

        Returns
        -------
        hikari.api.InteractionAutocompleteBuilder | None
            The autocomplete builder to send back to Discord, if using a REST client.
        """
        command = self.slash_commands.get(interaction.command_name)

        if command is None:
            logger.warning(f"Received autocomplete interaction for unknown command '{interaction.command_name}'.")
            return

        return await command._on_autocomplete(interaction)

    def include(self, command: CommandBase[te.Self, BuilderT]) -> CommandBase[te.Self, BuilderT]:
        """First-order decorator to add a command to this client.

        Parameters
        ----------
        command : CommandBase[te.Self, BuilderT]
            The command to add.
        """
        if command.plugin is not None:
            raise ValueError(f"Command '{command.name}' is already registered with plugin '{command.plugin.name}'.")

        if command.name in self.commands[command.command_type]:
            raise ValueError(f"Command '{command.name}' is already registered with this client.")

        command._client_include_hook(self)
        return command

    def include_slash_group(
        self,
        name: str,
        description: str = "No description provided.",
        *,
        guilds: hikari.UndefinedOr[t.Sequence[hikari.Snowflake]] = hikari.UNDEFINED,
        autodefer: bool | AutodeferMode = True,
        is_dm_enabled: bool = True,
        default_permissions: hikari.UndefinedOr[hikari.Permissions] = hikari.UNDEFINED,
        name_localizations: dict[hikari.Locale, str] | None = None,
        description_localizations: dict[hikari.Locale, str] | None = None,
        is_nsfw: bool = False,
    ) -> SlashGroup[te.Self]:
        """Add a new slash command group to this client.

        Parameters
        ----------
        name : str
            The name of the slash command group.
        description : str
            The description of the slash command group.
        guilds : hikari.UndefinedOr[t.Sequence[hikari.Snowflake]], optional
            The guilds to register the slash command group in, by default hikari.UNDEFINED
        autodefer : bool | AutodeferMode, optional
            If True, all commands in this group will automatically defer if it is taking longer than 2 seconds to respond.
            This can be overridden on a per-subcommand basis.
        is_dm_enabled : bool, optional
            Whether the slash command group is enabled in DMs, by default True
        default_permissions : hikari.UndefinedOr[hikari.Permissions], optional
            The default permissions for the slash command group, by default hikari.UNDEFINED
        name_localizations : dict[hikari.Locale, str], optional
            The name of the slash command group in different locales, by default None
        description_localizations : dict[hikari.Locale, str], optional
            The description of the slash command group in different locales, by default None
        is_nsfw : bool, optional
            Whether the slash command group is only usable in NSFW channels, by default False

        Returns
        -------
        SlashGroup[te.Self]
            The slash command group that was created.

        Usage
        -----
        ```py
        group = client.include_slash_group("Group", "A group of commands.")

        @group.include
        @arc.slash_subcommand(name="Command", description="A command.")
        async def cmd(ctx: arc.Context[arc.GatewayClient]) -> None:
            await ctx.respond("Hello!")
        ```
        """
        children: dict[str, SlashSubCommand[te.Self] | SlashSubGroup[te.Self]] = {}

        group = SlashGroup(
            name=name,
            description=description,
            children=children,
            guilds=guilds,
            autodefer=AutodeferMode(autodefer),
            is_dm_enabled=is_dm_enabled,
            default_permissions=default_permissions,
            name_localizations=name_localizations or {},
            description_localizations=description_localizations or {},
            is_nsfw=is_nsfw,
        )
        group._client_include_hook(self)
        return group

    def add_plugin(self, plugin: Plugin[te.Self]) -> None:
        """Add a plugin to this client.

        Parameters
        ----------
        plugin : Plugin[te.Self]
            The plugin to add.
        """
        plugin._client_include_hook(self)

    def remove_plugin(self, plugin: str | Plugin[te.Self]) -> None:
        """Remove a plugin from this client.

        Parameters
        ----------
        plugin : str | Plugin[te.Self]
            The plugin or name of the plugin to remove.

        Raises
        ------
        ValueError
            If there is no plugin with the given name.
        """
        if isinstance(plugin, Plugin):
            if plugin not in self.plugins.values():
                raise ValueError(f"Plugin '{plugin.name}' is not registered with this client.")
            return plugin._client_remove_hook()

        pg = self.plugins.get(plugin)

        if pg is None:
            raise ValueError(f"Plugin '{plugin}' is not registered with this client.")

        pg._client_remove_hook()

    def load_extension(self, path: str) -> te.Self:
        """Load a python module with path `path` as an extension.
        This will import the module, and call it's loader function.

        Parameters
        ----------
        path : str
            The path to the module to load.

        Returns
        -------
        te.Self
            The client for chaining calls.

        Raises
        ------
        ValueError
            If the module does not have a loader.

        Usage
        -----
        ```py
        client = arc.GatewayClient(...)
        client.load_extension("extension")

        # In extension.py

        plugin = arc.GatewayPlugin[arc.GatewayClient]("test_plugin")

        @arc.loader
        def loader(client: arc.GatewayClient) -> None:
            client.add_plugin(plugin)
        ```

        See Also
        --------
        - [`@arc.loader`][arc.extension.loader]
        - [`Client.load_extensions_from`][arc.client.Client.load_extensions_from]
        - [`Client.unload_extension`][arc.client.Client.unload_extension]
        """
        parents = path.split(".")
        name = parents.pop()

        pkg = ".".join(parents)

        if pkg:
            name = "." + name

        module = importlib.import_module(path, package=pkg)

        loader = getattr(module, "__arc_extension_loader__", None)

        if loader is None:
            raise ValueError(f"Module '{path}' does not have a loader.")

        self._loaded_extensions.append(path)
        loader(self)
        logger.info(f"Loaded extension: '{path}'")

        return self

    def load_extensions_from(self, dir_path: str | pathlib.Path, recursive: bool = False) -> te.Self:
        """Load all python modules in a directory as extensions.
        This will import the modules, and call their loader functions.

        Parameters
        ----------
        dir_path : str
            The path to the directory to load extensions from.
        recursive : bool, optional
            Whether to load extensions from subdirectories, by default False

        Returns
        -------
        te.Self
            The client for chaining calls.

        Raises
        ------
        ExtensionLoadError
            If `dir_path` does not exist or is not a directory.
        ExtensionLoadError
            If a module does not have a loader defined.
        """
        if isinstance(dir_path, str):
            dir_path = pathlib.Path(dir_path)

        try:
            dir_path.resolve().relative_to(pathlib.Path.cwd())
        except ValueError:
            raise ExtensionLoadError("dir_path must be relative to the current working directory.")

        if not dir_path.is_dir():
            raise ExtensionLoadError("dir_path must exist and be a directory.")

        globfunc = dir_path.rglob if recursive else dir_path.glob
        loaded = 0

        for file in globfunc(r"**/[!_]*.py"):
            module_path = ".".join(file.as_posix()[:-3].split("/"))
            self.load_extension(module_path)
            loaded += 1

        if loaded == 0:
            logger.warning(f"No extensions were found at '{dir_path}'.")

        return self

    def unload_extension(self, path: str) -> te.Self:
        """Unload a python module with path `path` as an extension.

        Parameters
        ----------
        path : str
            The path to the module to unload.

        Returns
        -------
        te.Self
            The client for chaining calls.

        Raises
        ------
        ExtensionUnloadError
            If the module does not have an unloader or is not loaded.
        """
        parents = path.split(".")
        name = parents.pop()

        pkg = ".".join(parents)

        if pkg:
            name = "." + name

        if path not in self._loaded_extensions:
            raise ExtensionUnloadError(f"Extension '{path}' is not loaded.")

        module = importlib.import_module(path, package=pkg)

        unloader = getattr(module, "__arc_extension_unloader__", None)

        if unloader is None:
            raise ExtensionUnloadError(f"Module '{path}' does not have an unloader.")

        unloader(self)
        self._loaded_extensions.remove(path)

        if module.__name__ in sys.modules:
            del sys.modules[module.__name__]

        return self

    def set_type_dependency(self, type_: t.Type[T], instance: T) -> None:
        """Set a type dependency for this client. This can then be injected into all arc callbacks.

        Parameters
        ----------
        type_ : t.Type[T]
            The type of the dependency.
        instance : T
            The instance of the dependency.

        Usage
        -----

        ```py
        class MyDependency:
            def __init__(self, value: str):
                self.value = value

        client.set_type_dependency(MyDependency, MyDependency("Hello!"))

        @client.include
        @arc.slash_command("cmd", "A command.")
        async def cmd(ctx: arc.Context[arc.GatewayClient], dep: arc.Injected[MyDependency]) -> None:
            await ctx.respond(dep.value)
        ```

        See Also
        --------
        - [`Client.get_type_dependency`][arc.client.Client.get_type_dependency]
            A method to get dependencies for the client.

        - [`Client.inject_dependencies`][arc.client.Client.inject_dependencies]
            A decorator to inject dependencies into arbitrary functions.
        """
        self._injector.set_type_dependency(type_, instance)

    def get_type_dependency(self, type_: t.Type[T]) -> hikari.UndefinedOr[T]:
        """Get a type dependency for this client.

        Parameters
        ----------
        type_ : t.Type[T]
            The type of the dependency.

        Returns
        -------
        hikari.UndefinedOr[T]
            The instance of the dependency, if it exists.
        """
        return self._injector.get_type_dependency(type_, default=hikari.UNDEFINED)

    def inject_dependencies(self, func: t.Callable[P, T]) -> t.Callable[P, T]:
        """First order decorator to inject dependencies into the decorated function.

        !!! note
            Command callbacks are automatically injected with dependencies,
            thus this decorator is not needed for them.

        Usage
        -----
        ```py
        class MyDependency:
            def __init__(self, value: str):
                self.value = value

        client.set_type_dependency(MyDependency, MyDependency("Hello!"))

        @client.inject_dependencies
        def my_func(dep: MyDependency = arc.inject()) -> None:
            print(dep.value)

        my_func() # Prints "Hello!"
        ```

        See Also
        --------
        - [`Client.set_type_dependency`][arc.client.Client.set_type_dependency]
            A method to set dependencies for the client.
        """
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def decorator_async(*args: P.args, **kwargs: P.kwargs) -> T:
                return await self.injector.call_with_async_di(func, *args, **kwargs)

            return decorator_async  # pyright: ignore reportGeneralTypeIssues
        else:

            @functools.wraps(func)
            def decorator(*args: P.args, **kwargs: P.kwargs) -> T:
                return self.injector.call_with_di(func, *args, **kwargs)

            return decorator

    async def resync_commands(self) -> None:
        """Synchronize the commands registered in this client with Discord.

        !!! warning
            Calling this is expensive, and should only be done when absolutely necessary.
            The client automatically syncs commands on startup, unless the `autosync` parameter
            is set to `False` when creating the client.

        Raises
        ------
        RuntimeError
            If `Client.application` is `None`.
            This usually only happens if `Client.resync_commands` is called before `Client.on_startup`.
        """
        await _sync_commands(self)

    async def purge_all_commands(self, guild: hikari.SnowflakeishOr[hikari.PartialGuild] | None = None) -> None:
        """Purge all commands registered on Discord. This can be used to clean up commands.

        Parameters
        ----------
        guild : hikari.SnowflakeishOr[hikari.PartialGuild] | None, optional
            The guild to purge commands from, by default None
            If a `guild` is not provided, this will purge global commands.

        !!! warning
            This will remove all commands registered on Discord, **including commands not registered by this client**.

        Raises
        ------
        RuntimeError
            If `Client.application` is `None`.
            This usually only happens if `Client.purge_all_commands` is called before `Client.on_startup`.
        """
        if self.application is None:
            raise RuntimeError(f"Cannot purge commands before '{type(self).__name__}.application' is fetched.")

        if guild is not None:
            guild_id = hikari.Snowflake(guild)
            await self.rest.set_application_commands(self.application, [], guild_id)
        else:
            await self.rest.set_application_commands(self.application, [])


class GatewayClient(Client[hikari.GatewayBotAware]):
    """A base class for an arc client with `hikari.GatewayBotAware` support.
    If you want to use a `hikari.RESTBotAware`, use `RESTClient` instead.

    Parameters
    ----------
    app : hikari.GatewayBot
        The application this client is for.
    default_enabled_guilds : t.Sequence[hikari.Snowflake] | None, optional
        The guilds that slash commands will be registered in by default, by default None
    autosync : bool, optional
        Whether to automatically sync commands on startup, by default True

    Usage
    -----
    ```py
    import hikari
    import arc

    bot = hikari.GatewayBot("TOKEN")
    client = arc.GatewayClient(bot)

    ...
    ```
    """

    __slots__: t.Sequence[str] = ()

    def __init__(
        self,
        app: hikari.GatewayBotAware,
        *,
        default_enabled_guilds: t.Sequence[hikari.Snowflake] | None = None,
        autosync: bool = True,
    ) -> None:
        super().__init__(app, default_enabled_guilds=default_enabled_guilds, autosync=autosync)
        self.app.event_manager.subscribe(hikari.StartedEvent, self._on_gatewaybot_startup)
        self.app.event_manager.subscribe(hikari.StoppingEvent, self._on_gatewaybot_shutdown)
        self.app.event_manager.subscribe(hikari.InteractionCreateEvent, self._on_gatewaybot_interaction_create)

    @property
    def is_rest(self) -> bool:
        """Whether this client is a REST client."""
        return False

    @property
    def cache(self) -> hikari.api.Cache:
        """The cache for this client."""
        return self.app.cache

    async def _on_gatewaybot_startup(self, event: hikari.StartedEvent) -> None:
        await self._on_startup()

    async def _on_gatewaybot_shutdown(self, event: hikari.StoppingEvent) -> None:
        await self._on_shutdown()

    async def _on_gatewaybot_interaction_create(self, event: hikari.InteractionCreateEvent) -> None:
        if isinstance(event.interaction, hikari.CommandInteraction):
            await self.on_command_interaction(event.interaction)
        elif isinstance(event.interaction, hikari.AutocompleteInteraction):
            await self.on_autocomplete_interaction(event.interaction)

    async def _on_error(self, ctx: Context[te.Self], exception: Exception) -> None:
        if not self.app.event_manager.get_listeners(CommandErrorEvent):
            return await super()._on_error(ctx, exception)

        self.app.event_manager.dispatch(CommandErrorEvent(self, ctx, exception))

    def listen(self, *event_types: t.Type[EventT]) -> t.Callable[[EventCallbackT[EventT]], EventCallbackT[EventT]]:
        """Generate a decorator to subscribe a callback to an event type.

        This is a second-order decorator.

        Parameters
        ----------
        *event_types : t.Type[EventT] | None
            The event types to subscribe to. The implementation may allow this
            to be undefined. If this is the case, the event type will be inferred
            instead from the type hints on the function signature.

            `EventT` must be a subclass of `hikari.events.base_events.Event`.

        Returns
        -------
        t.Callable[[EventT], EventT]
            A decorator for a coroutine function that passes it to
            `EventManager.subscribe` before returning the function
            reference.
        """
        return self.app.event_manager.listen(*event_types)


class RESTClient(Client[hikari.RESTBotAware]):
    """A base class for an arc client with `hikari.RESTBotAware` support.
    If you want to use `hikari.GatewayBotAware`, use `GatewayClient` instead.

    Parameters
    ----------
    app : hikari.RESTBot
        The application this client is for.
    default_enabled_guilds : t.Sequence[hikari.Snowflake] | None, optional
        The guilds that slash commands will be registered in by default, by default None
    autosync : bool, optional
        Whether to automatically sync commands on startup, by default True


    Usage
    -----
    ```py
    import hikari
    import arc

    bot = hikari.RESTBot("TOKEN")
    client = arc.RESTClient(bot)

    ...
    ```

    !!! warning
        The client will take over the `hikari.CommandInteraction` and `hikari.AutocompleteInteraction` listeners of the passed bot.
    """

    __slots__: t.Sequence[str] = ()

    def __init__(
        self,
        app: hikari.RESTBotAware,
        *,
        default_enabled_guilds: t.Sequence[hikari.Snowflake] | None = None,
        autosync: bool = True,
    ) -> None:
        super().__init__(app, default_enabled_guilds=default_enabled_guilds, autosync=autosync)
        self.app.add_startup_callback(self._on_restbot_startup)
        self.app.add_shutdown_callback(self._on_restbot_shutdown)
        self.app.interaction_server.set_listener(
            hikari.CommandInteraction, self._on_restbot_interaction_create, replace=True
        )
        self.app.interaction_server.set_listener(
            hikari.AutocompleteInteraction, self._on_restbot_autocomplete_interaction_create, replace=True
        )

    @property
    def is_rest(self) -> bool:
        """Whether this client is a REST client."""
        return True

    async def _on_restbot_startup(self, bot: hikari.RESTBotAware) -> None:
        await self._on_startup()
        await self.on_restbot_startup(bot)

    async def on_restbot_startup(self, bot: hikari.RESTBotAware) -> None:
        """A function that is called when the client is started.

        Parameters
        ----------
        bot : hikari.RESTBotAware
            The bot that was started.
        """

    async def _on_restbot_shutdown(self, bot: hikari.RESTBotAware) -> None:
        await self._on_shutdown()
        await self.on_restbot_shutdown(bot)

    async def on_restbot_shutdown(self, bot: hikari.RESTBotAware) -> None:
        """A function that is called when the client is shut down.

        Parameters
        ----------
        bot : hikari.RESTBotAware
            The bot that was shut down.
        """

    async def _on_restbot_interaction_create(self, interaction: hikari.CommandInteraction) -> ResponseBuilderT:
        builder = await self.on_command_interaction(interaction)
        assert builder is not None
        return builder

    async def _on_restbot_autocomplete_interaction_create(
        self, interaction: hikari.AutocompleteInteraction
    ) -> hikari.api.InteractionAutocompleteBuilder:
        builder = await self.on_autocomplete_interaction(interaction)
        assert builder is not None
        return builder


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import asyncio
import typing as t

import attr
import hikari

from ..abc import HasErrorHandler
from ..context import AutocompleteData, Context
from ..errors import AutocompleteError, CommandInvokeError
from ..internal.sigparse import parse_function_signature
from ..internal.types import ClientT, CommandCallbackT, ResponseBuilderT, SlashCommandLike
from .base import AutodeferMode, CallableCommandBase, CommandBase
from .option import OptionBase, OptionWithChoices

if t.TYPE_CHECKING:
    from asyncio.futures import Future

    from ..plugin import Plugin
    from .base import CallableCommandProto
    from .option import CommandOptionBase

__all__ = (
    "SlashCommandLike",
    "SlashCommand",
    "SlashGroup",
    "SlashSubCommand",
    "SlashSubGroup",
    "slash_command",
    "slash_subcommand",
)


def _resolve_options(
    local_options: t.MutableMapping[str, CommandOptionBase[ClientT, t.Any, t.Any]],
    incoming_options: t.Sequence[hikari.CommandInteractionOption],
    resolved: hikari.ResolvedOptionData | None,
) -> dict[str, t.Any]:
    """Resolve the options into kwargs for the callback.

    Parameters
    ----------
    local_options : t.MutableMapping[str, Option[t.Any, t.Any]]
        The options of the locally stored command.
    incoming_options : t.Sequence[hikari.CommandInteractionOption]
        The options of the interaction.
    resolved : hikari.ResolvedOptionData
        The resolved option data of the interaction.

    Returns
    -------
    dict[str, Any]
        The resolved options as kwargs, ready to be passed to the callback.
    """
    option_kwargs: dict[str, t.Any] = {}

    for arg_name, opt in local_options.items():
        inter_opt = next((o for o in incoming_options if o.name == opt.name), None)

        if inter_opt is None:
            continue

        if isinstance(inter_opt.value, hikari.Snowflake) and resolved:
            match inter_opt.type:
                case hikari.OptionType.USER:
                    value = resolved.members.get(inter_opt.value) or resolved.users[inter_opt.value]
                case hikari.OptionType.ATTACHMENT:
                    value = resolved.attachments[inter_opt.value]
                case hikari.OptionType.CHANNEL:
                    value = resolved.channels[inter_opt.value]
                case hikari.OptionType.ROLE:
                    value = resolved.roles[inter_opt.value]
                case hikari.OptionType.MENTIONABLE:
                    value = (
                        resolved.members.get(inter_opt.value)
                        or resolved.users.get(inter_opt.value)
                        or resolved.roles[inter_opt.value]
                    )
                case _:
                    raise ValueError(f"Unexpected option type '{inter_opt.type}.'")

            option_kwargs[arg_name] = value

        elif isinstance(inter_opt.value, hikari.Snowflake):
            raise ValueError(f"Missing resolved option data for '{inter_opt.name}'.")
        else:
            option_kwargs[arg_name] = inter_opt.value

    return option_kwargs


def _choices_to_builders(
    choices: t.Sequence[hikari.api.AutocompleteChoiceBuilder] | t.Sequence[t.Any],
) -> t.Sequence[hikari.api.AutocompleteChoiceBuilder]:
    """Convert a sequence of choices to a sequence of choice builders."""
    return [
        (
            hikari.impl.AutocompleteChoiceBuilder(str(e), e)
            if not isinstance(e, hikari.api.AutocompleteChoiceBuilder)
            else e
        )
        for e in choices
    ]


@attr.define(slots=True, kw_only=True)
class SlashCommand(CallableCommandBase[ClientT, hikari.api.SlashCommandBuilder]):
    """A slash command outside of any group."""

    description: str = "No description provided."
    """The description of this slash command."""

    description_localizations: dict[hikari.Locale, str] = attr.field(factory=dict)
    """The localizations for this command's description."""

    options: dict[str, CommandOptionBase[t.Any, ClientT, t.Any]] = attr.field(factory=dict)
    """The options of this slash command."""

    @property
    def command_type(self) -> hikari.CommandType:
        return hikari.CommandType.SLASH

    @property
    def qualified_name(self) -> t.Sequence[str]:
        return (self.name,)

    def _get_context(
        self, interaction: hikari.CommandInteraction, command: CallableCommandProto[ClientT]
    ) -> Context[ClientT]:
        assert self.client is not None

        if interaction.command_type is not hikari.CommandType.SLASH:
            raise ValueError(f"Expected slash command, got {interaction.command_type}")

        return Context(self.client, command, interaction)

    def _to_dict(self) -> dict[str, t.Any]:
        sorted_options = sorted(self.options.values(), key=lambda option: option.is_required, reverse=True)
        return {
            **super()._to_dict(),
            "description": self.description,
            "description_localizations": self.description_localizations,
            "options": [option.to_command_option() for option in sorted_options],
        }

    def _build(self) -> hikari.api.SlashCommandBuilder:
        return hikari.impl.SlashCommandBuilder(
            name=self.name,
            description=self.description,
            options=[option.to_command_option() for option in self.options.values()],
            default_member_permissions=self.default_permissions,
            is_dm_enabled=self.is_dm_enabled,
            is_nsfw=self.is_nsfw,
            name_localizations=self.name_localizations,  # pyright: ignore reportGeneralTypeIssues
            description_localizations=self.description_localizations,  # pyright: ignore reportGeneralTypeIssues
        )

    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> Future[ResponseBuilderT] | None:
        if interaction.options:
            return await super().invoke(
                interaction,
                *args,
                **{**kwargs, **_resolve_options(self.options, interaction.options, interaction.resolved)},
            )
        else:
            return await super().invoke(interaction, *args, **kwargs)

    async def _on_autocomplete(
        self, interaction: hikari.AutocompleteInteraction
    ) -> hikari.api.InteractionAutocompleteBuilder | None:
        opt = next((o for o in interaction.options if o.name in self.options and o.is_focused), None)

        if opt is None:
            raise ValueError(f"Slash command got unknown option to autocomplete: '{interaction.options[0]}'.")

        local_opt = self.options[opt.name]

        if not isinstance(local_opt, OptionWithChoices) or not local_opt.autocomplete_with:
            raise ValueError(
                f"Slash option got autocomplete interaction without autocomplete callback: '{local_opt.name}'."
            )

        choices = _choices_to_builders(
            await self.client.injector.call_with_async_di(
                local_opt.autocomplete_with,  # pyright: ignore reportGeneralTypeIssues
                AutocompleteData(
                    interaction=interaction, options=interaction.options, client=self.client, command=self
                ),
            )
        )

        if self.client.is_rest:
            return interaction.build_response(choices)

        await interaction.create_response(choices)


@attr.define(slots=True, kw_only=True)
class SlashGroup(CommandBase[ClientT, hikari.api.SlashCommandBuilder]):
    """A group for slash subcommands and subgroups."""

    children: dict[str, SlashSubCommand[ClientT] | SlashSubGroup[ClientT]] = attr.field(factory=dict)
    """Subcommands and subgroups that belong to this group."""

    description: str = "No description provided."
    """The description of this slash group."""

    description_localizations: dict[hikari.Locale, str] = attr.field(factory=dict)
    """The localizations for this group's description."""

    _invoke_task: asyncio.Task[t.Any] | None = attr.field(init=False, default=None)

    @property
    def command_type(self) -> hikari.CommandType:
        return hikari.CommandType.SLASH

    @property
    def qualified_name(self) -> t.Sequence[str]:
        return (self.name,)

    def _to_dict(self) -> dict[str, t.Any]:
        return {
            **super()._to_dict(),
            "description": self.description,
            "description_localizations": self.description_localizations,
            "options": [subcmd.to_command_option() for subcmd in self.children.values()],
        }

    def _get_context(
        self, interaction: hikari.CommandInteraction, command: CallableCommandProto[ClientT]
    ) -> Context[ClientT]:
        assert self.client is not None

        if interaction.command_type is not hikari.CommandType.SLASH:
            raise ValueError(f"Expected slash command, got {interaction.command_type}")

        return Context(self.client, command, interaction)

    def _build(self) -> hikari.api.SlashCommandBuilder:
        return hikari.impl.SlashCommandBuilder(
            name=self.name,
            description=self.description,
            options=[subcmd.to_command_option() for subcmd in self.children.values()],
            default_member_permissions=self.default_permissions,
            is_dm_enabled=self.is_dm_enabled,
            is_nsfw=self.is_nsfw,
            name_localizations=self.name_localizations,  # pyright: ignore reportGeneralTypeIssues
            description_localizations=self.description_localizations,  # pyright: ignore reportGeneralTypeIssues
        )

    async def _invoke_subcmd(
        self,
        subcommand: SlashSubCommand[ClientT],
        interaction: hikari.CommandInteraction,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> asyncio.Future[ResponseBuilderT] | None:
        """Invoke a subcommand."""
        ctx = self._get_context(interaction, subcommand)

        if (autodefer := subcommand._resolve_autodefer()) and autodefer.should_autodefer:
            ctx._start_autodefer(autodefer)

        self._invoke_task = asyncio.create_task(self._handle_callback(subcommand, ctx, *args, **kwargs))
        if self.client.is_rest:
            return ctx._resp_builder

    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> Future[ResponseBuilderT] | None:
        if interaction.options is None:
            raise CommandInvokeError("Cannot invoke slash group with empty options.")

        # Get first-order subcommand
        sub = next((o for o in interaction.options if o.name in self.children), None)

        if sub is None:
            raise CommandInvokeError(f"Slash group got unknown subcommand: '{interaction.options[0]}'.")

        subcmd = self.children[sub.name]

        # Invoke it if it has no options
        if sub.options is None:
            if not isinstance(subcmd, SlashSubCommand):
                raise CommandInvokeError(f"Slash group got subgroup without options: '{subcmd.name}'.")
            return await self._invoke_subcmd(subcmd, interaction, *args, **kwargs)

        # Resolve options and invoke if it does
        if isinstance(subcmd, SlashSubCommand):
            res = _resolve_options(subcmd.options, sub.options, interaction.resolved)
            return await self._invoke_subcmd(subcmd, interaction, *args, **{**kwargs, **res})

        # Get second-order subcommand
        subsub = next((o for o in sub.options if o.name in subcmd.children), None)

        if subsub is None:
            raise CommandInvokeError(f"Slash group got unknown subcommand: '{sub.options[0]}'.")

        subsubcmd = subcmd.children[subsub.name]

        # Invoke it if it has no options
        if subsub.options is None:
            return await self._invoke_subcmd(subsubcmd, interaction, *args, **kwargs)

        # Resolve options and invoke if it does
        res = _resolve_options(subsubcmd.options, subsub.options, interaction.resolved)
        return await self._invoke_subcmd(subsubcmd, interaction, *args, **{**kwargs, **res})

    async def _on_autocomplete(
        self, interaction: hikari.AutocompleteInteraction
    ) -> hikari.api.InteractionAutocompleteBuilder | None:
        # First-order subcommand
        sub = next((o for o in interaction.options if o.name in self.children), None)

        if sub is None:
            raise AutocompleteError(f"Slash group got unknown subcommand to autocomplete: '{interaction.options[0]}'.")

        subcmd = self.children[sub.name]

        if sub.options is None:
            raise AutocompleteError(f"Slash group got subcommand without options: '{subcmd.name}'.")

        # If it is a first-order subcommand, get the option
        if isinstance(subcmd, SlashSubCommand):
            opts = sub.options
            opt = next(o for o in sub.options if o.name in subcmd.options and o.is_focused)
            local_opt = subcmd.options[opt.name]
        else:
            # Otherwise continue the conga-line
            subsub = next((o for o in sub.options if o.name in subcmd.children), None)

            if subsub is None:
                raise AutocompleteError(f"Slash subgroup got unknown subcommand to autocomplete: '{sub.options[0]}'.")

            if subsub.options is None:
                raise AutocompleteError(f"Slash subgroup got subcommand without options: '{subsub.name}'.")

            subsubcmd = subcmd.children[subsub.name]
            opts = subsub.options
            opt = next(o for o in subsub.options if o.name in subsubcmd.options and o.is_focused)
            local_opt = subsubcmd.options[opt.name]

        if not isinstance(local_opt, OptionWithChoices) or not local_opt.autocomplete_with:
            raise AutocompleteError(
                f"Slash option got autocomplete interaction without autocomplete callback: '{local_opt.name}'."
            )

        choices = _choices_to_builders(
            await self.client.injector.call_with_async_di(
                local_opt.autocomplete_with,  # pyright: ignore reportGeneralTypeIssues
                AutocompleteData(interaction=interaction, options=opts, client=self.client, command=self),
            )
        )

        if self.client.is_rest:
            return interaction.build_response(choices)

        await interaction.create_response(choices)

    def include(self, command: SlashSubCommand[ClientT]) -> SlashSubCommand[ClientT]:
        """First-order decorator to add a subcommand to this group."""
        command.parent = self
        self.children[command.name] = command
        return command

    def include_subgroup(
        self,
        name: str,
        description: str = "No description provided.",
        *,
        autodefer: AutodeferMode | bool | hikari.UndefinedType = hikari.UNDEFINED,
        name_localizations: dict[hikari.Locale, str] | None = None,
        description_localizations: dict[hikari.Locale, str] | None = None,
    ) -> SlashSubGroup[ClientT]:
        """Create a subgroup and add it to this group.

        Parameters
        ----------
        name : str
            The name of the subgroup.
        description : str, optional
            The description of the subgroup, by default "No description provided."
        autodefer : bool | AutodeferMode | hikari.UndefinedType, optional
            If True, all commands in this subgroup will automatically defer if it is taking longer than 2 seconds to respond.
            If not provided, then this setting will be inherited from the parent.
        name_localizations : dict[hikari.Locale, str] | None, optional
            Localizations for the name of the subgroup, by default None
        description_localizations : dict[hikari.Locale, str] | None, optional
            Localizations for the description of the subgroup, by default None
        """
        group: SlashSubGroup[ClientT] = SlashSubGroup(
            name=name,
            description=description,
            autodefer=AutodeferMode(autodefer) if autodefer else hikari.UNDEFINED,
            name_localizations=name_localizations or {},
            description_localizations=description_localizations or {},
            parent=self,
        )
        self.children[name] = group
        return group


@attr.define(slots=True, kw_only=True)
class SlashSubGroup(OptionBase[ClientT], HasErrorHandler[ClientT]):
    """A subgroup of a slash command group."""

    parent: SlashGroup[ClientT] | None = None
    """The parent group of this subgroup."""

    children: dict[str, SlashSubCommand[ClientT]] = attr.field(factory=dict)
    """Subcommands that belong to this subgroup."""

    autodefer: AutodeferMode | hikari.UndefinedType = hikari.UNDEFINED
    """If True, this subcommand will automatically defer if it is taking longer than 2 seconds to respond.
    If undefined, then it will be inherited from the parent.
    """

    _plugin: Plugin[ClientT] | None = attr.field(default=None, init=False)

    @property
    def option_type(self) -> hikari.OptionType:
        return hikari.OptionType.SUB_COMMAND_GROUP

    @property
    def command_type(self) -> hikari.CommandType:
        return hikari.CommandType.SLASH

    @property
    def qualified_name(self) -> t.Sequence[str]:
        if self.parent is None:
            raise ValueError("Cannot get qualified name of subgroup without parent.")

        return (self.parent.name, self.name)

    @property
    def client(self) -> ClientT:
        """The client that includes this subgroup."""
        if self.parent is None:
            raise ValueError("Cannot get client of subgroup without parent.")

        return self.parent.client

    @property
    def plugin(self) -> Plugin[ClientT] | None:
        """The plugin that includes this subgroup."""
        return self._plugin

    def _to_dict(self) -> dict[str, t.Any]:
        return {
            **super()._to_dict(),
            "options": [subcommand.to_command_option() for subcommand in self.children.values()],
        }

    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        try:
            if self.error_handler:
                await self.error_handler(ctx, exc)
            else:
                raise exc
        except Exception as e:
            assert self.parent is not None
            await self.parent._handle_exception(ctx, e)

    def include(self, command: SlashSubCommand[ClientT]) -> SlashSubCommand[ClientT]:
        """First-order decorator to add a subcommand to this group."""
        command.parent = self
        self.children[command.name] = command
        return command


@attr.define(slots=True, kw_only=True)
class SlashSubCommand(OptionBase[ClientT], HasErrorHandler[ClientT]):
    """A subcommand of a slash command group."""

    parent: SlashGroup[ClientT] | SlashSubGroup[ClientT] | None = None
    """The parent group of this subcommand."""

    callback: CommandCallbackT[ClientT]
    """The callback that will be invoked when this subcommand is invoked."""

    options: t.MutableMapping[str, CommandOptionBase[ClientT, t.Any, t.Any]] = attr.field(factory=dict)
    """The options of this subcommand."""

    autodefer: AutodeferMode | hikari.UndefinedType = hikari.UNDEFINED
    """If True, this subcommand will automatically defer if it is taking longer than 2 seconds to respond.
    If undefined, then it will be inherited from the parent.
    """

    _invoke_task: asyncio.Task[t.Any] | None = attr.field(default=None, init=False)

    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        try:
            if self.error_handler:
                await self.error_handler(ctx, exc)
            else:
                raise exc
        except Exception as e:
            assert self.parent is not None
            await self.parent._handle_exception(ctx, e)

    @property
    def qualified_name(self) -> t.Sequence[str]:
        if self.parent is None:
            raise ValueError("Cannot get qualified name of subcommand without parent.")

        if isinstance(self.parent, SlashSubGroup):
            if self.parent.parent is None:
                raise ValueError("Cannot get qualified name of subcommand without parent.")

            return (self.parent.parent.name, self.parent.name, self.name)

        return (self.parent.name, self.name)

    @property
    def root(self) -> SlashGroup[ClientT]:
        """The root group of this subcommand."""
        if self.parent is None:
            raise ValueError("Cannot get root of subcommand without parent.")

        if isinstance(self.parent, SlashSubGroup):
            if self.parent.parent is None:
                raise ValueError("Cannot get root of subcommand without parent.")

            return self.parent.parent

        return self.parent

    @property
    def command_type(self) -> hikari.CommandType:
        return hikari.CommandType.SLASH

    @property
    def option_type(self) -> hikari.OptionType:
        return hikari.OptionType.SUB_COMMAND

    @property
    def client(self) -> ClientT:
        """The client that includes this subcommand."""
        return self.root.client

    @property
    def plugin(self) -> Plugin[ClientT] | None:
        """The plugin that includes this subcommand."""
        return self.root.plugin

    def _resolve_autodefer(self) -> AutodeferMode:
        """Resolve autodefer for this subcommand."""
        if self.autodefer is not hikari.UNDEFINED:
            return self.autodefer

        if self.parent is None:
            return AutodeferMode.OFF

        if isinstance(self.parent, SlashSubGroup):
            if self.parent.autodefer is not hikari.UNDEFINED:
                return self.parent.autodefer

            if self.parent.parent is None:
                return AutodeferMode.OFF

            return self.parent.parent.autodefer

        return self.parent.autodefer

    async def __call__(self, ctx: Context[ClientT], *args: t.Any, **kwargs: t.Any) -> None:
        """Invoke this subcommand with the given context.

        Parameters
        ----------
        ctx : Context
            The context to invoke this subcommand with.
        args: list[t.Any]
            The positional arguments to pass to the callback.
        kwargs : dict[str, Any]
            The keyword arguments to pass to the callback.
        """
        await self.callback(ctx, *args, **kwargs)

    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> asyncio.Future[ResponseBuilderT] | None:
        return await self.root._invoke_subcmd(self, interaction, *args, **kwargs)

    def _to_dict(self) -> dict[str, t.Any]:
        sorted_options = sorted(self.options.values(), key=lambda option: option.is_required, reverse=True)
        return {**super()._to_dict(), "options": [option.to_command_option() for option in sorted_options]}


def slash_command(
    name: str,
    description: str = "No description provided.",
    *,
    guilds: t.Sequence[hikari.SnowflakeishOr[hikari.PartialGuild]] | None = None,
    is_dm_enabled: bool = True,
    is_nsfw: bool = False,
    autodefer: bool | AutodeferMode = True,
    default_permissions: hikari.UndefinedOr[hikari.Permissions] = hikari.UNDEFINED,
    name_localizations: dict[hikari.Locale, str] | None = None,
    description_localizations: dict[hikari.Locale, str] | None = None,
) -> t.Callable[[t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]]], SlashCommand[ClientT]]:
    """A decorator that creates a slash command.

    Parameters
    ----------
    name : str
        The name of the slash command.
    description : str, optional
        The description of the command, by default "No description provided."
    guilds : t.Sequence[hikari.SnowflakeishOr[hikari.PartialGuild]] | None, optional
        The guilds this command should be enabled in, if left as None, the command is global, by default None
    is_dm_enabled : bool, optional
        If True, the command is usable in direct messages, by default True
    is_nsfw : bool, optional
        If True, the command is only usable in NSFW channels, by default False
    autodefer : bool | AutodeferMode, optional
        If True, this command will be automatically deferred if it takes longer than 2 seconds to respond, by default True
    default_permissions : hikari.UndefinedOr[hikari.Permissions], optional
        The default permissions required to use this command, these can be overriden by guild admins, by default hikari.UNDEFINED
    name_localizations : dict[hikari.Locale, str] | None, optional
        Localizations for the name of this command, by default None
    description_localizations : dict[hikari.Locale, str] | None, optional
        Localizations for the description of this command, by default None

    Returns
    -------
    t.Callable[[t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]]], SlashCommand[ClientT]]
        The decorated slash command.

    Usage
    -----
    ```py
    @client.include
    @arc.slash_command(name="hi", description="Say hi!")
    async def hi_slash(
        ctx: arc.Context[arc.GatewayClient],
        user: arc.Option[hikari.User, arc.UserParams(description="The user to say hi to.")]
    ) -> None:
        await ctx.respond(f"Hey {user.mention}!")
    ```
    """

    def decorator(func: t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]]) -> SlashCommand[ClientT]:
        guild_ids = [hikari.Snowflake(guild) for guild in guilds] if guilds else []
        options = parse_function_signature(func)

        return SlashCommand(
            callback=func,
            options=options,
            autodefer=AutodeferMode(autodefer),
            name=name,
            description=description,
            name_localizations=name_localizations or {},
            description_localizations=description_localizations or {},
            default_permissions=default_permissions,
            guilds=guild_ids,
            is_dm_enabled=is_dm_enabled,
            is_nsfw=is_nsfw,
        )

    return decorator


def slash_subcommand(
    name: str,
    description: str = "No description provided.",
    *,
    autodefer: bool | AutodeferMode | hikari.UndefinedType = hikari.UNDEFINED,
    name_localizations: dict[hikari.Locale, str] | None = None,
    description_localizations: dict[hikari.Locale, str] | None = None,
) -> t.Callable[[t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]]], SlashSubCommand[ClientT]]:
    """A decorator that creates a slash sub command. It should be included in a slash command group.

    Parameters
    ----------
    name : str
        The name of the slash command.
    description : str, optional
        The description of the command, by default "No description provided."
    autodefer : bool | AutodeferMode | hikari.UndefinedType, optional
        If True, this command will be automatically deferred if it takes longer than 2 seconds to respond, by default True
        If undefined, then this setting will be be inherited from the parent
    name_localizations : dict[hikari.Locale, str] | None, optional
        Localizations for the name of this command, by default None
    description_localizations : dict[hikari.Locale, str] | None, optional
        Localizations for the description of this command, by default None

    Returns
    -------
    t.Callable[[t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]]], SlashCommand[ClientT]]
        The decorated slash command.

    Usage
    -----
    ```py
    group = client.include_group("group", "A group of slash commands.")

    @group.include
    @arc.slash_subcommand(name="hi", description="Say hi!")
    async def hi_slashsub(
        ctx: arc.Context[arc.GatewayClient],
        user: arc.Option[hikari.User, arc.UserParams(description="The user to say hi to.")]
    ) -> None:
        await ctx.respond(f"Hey {user.mention}!")
    ```
    """

    def decorator(
        func: t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]],
    ) -> SlashSubCommand[ClientT]:
        options = parse_function_signature(func)

        return SlashSubCommand(
            callback=func,
            options=options,
            name=name,
            autodefer=AutodeferMode(autodefer) if autodefer else hikari.UNDEFINED,
            description=description,
            name_localizations=name_localizations or {},
            description_localizations=description_localizations or {},
        )

    return decorator


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc as arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(bot)


@client.include
@arc.slash_command("test", default_permissions=hikari.Permissions.ADMINISTRATOR)
async def my_command(
    ctx: arc.Context[arc.GatewayClient],
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
    c: arc.Option[float | None, arc.FloatParams(description="baz", max=50.0)],
    d: arc.Option[hikari.GuildTextChannel | hikari.GuildNewsChannel | None, arc.ChannelParams(description="qux")],
    e: arc.Option[hikari.GuildChannel | None, arc.ChannelParams(description="quux")],
    f: arc.Option[hikari.Role | hikari.User, arc.MentionableParams(description="quuz")],
    g: arc.Option[hikari.Attachment | None, arc.AttachmentParams(description="among us")],
) -> None:
    pass


group = client.include_slash_group(
    "my_group", "My group description", default_permissions=hikari.Permissions.ADMINISTRATOR
)

subgroup = group.include_subgroup("my_subgroup", "My subgroup description")


@group.include
@arc.slash_subcommand("test_subcommand", "My subcommand description")
async def my_subcommand(
    ctx: arc.Context[arc.GatewayClient],
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
) -> None:
    pass


@subgroup.include
@arc.slash_subcommand("test_subsubcommand", "My subsubcommand description")
async def my_subsubcommand(
    ctx: arc.Context[arc.GatewayClient],
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
) -> None:
    pass


def test_my_command() -> None:
    assert len(client.slash_commands) == 2

    command = client.slash_commands["test"]

    assert command.client is client

    assert command.name == "test"
    assert command.description == "No description provided."
    assert command.default_permissions == hikari.Permissions.ADMINISTRATOR
    assert command.is_dm_enabled is True
    assert command.is_nsfw is False
    assert command.name_localizations == {}
    assert command.description_localizations == {}

    assert isinstance(command, arc.SlashCommand)
    options = command.options

    assert isinstance(options["a"], arc.IntOption)
    assert options["a"].name == "a"
    assert options["a"].description == "foo"
    assert options["a"].is_required
    assert options["a"].min == 10
    assert options["a"].max is None

    assert isinstance(options["b"], arc.StrOption)
    assert options["b"].name == "b"
    assert options["b"].description == "bar"
    assert options["b"].is_required
    assert options["b"].min_length == 100
    assert options["b"].max_length is None

    assert isinstance(options["c"], arc.FloatOption)
    assert options["c"].name == "c"
    assert options["c"].description == "baz"
    assert not options["c"].is_required
    assert options["c"].min is None
    assert options["c"].max == 50.0

    assert isinstance(options["d"], arc.ChannelOption)
    assert options["d"].name == "d"
    assert options["d"].description == "qux"
    assert not options["d"].is_required
    assert options["d"].channel_types is not None
    assert set(options["d"].channel_types) == {hikari.ChannelType.GUILD_TEXT, hikari.ChannelType.GUILD_NEWS}

    assert isinstance(options["e"], arc.ChannelOption)
    assert options["e"].name == "e"
    assert options["e"].description == "quux"
    assert not options["e"].is_required
    assert options["e"].channel_types is not None
    assert set(options["e"].channel_types) == {
        hikari.ChannelType.GUILD_TEXT,
        hikari.ChannelType.GUILD_VOICE,
        hikari.ChannelType.GUILD_CATEGORY,
        hikari.ChannelType.GUILD_NEWS,
        hikari.ChannelType.GUILD_FORUM,
        hikari.ChannelType.GUILD_NEWS_THREAD,
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
        hikari.ChannelType.GUILD_STAGE,
    }

    assert isinstance(options["f"], arc.MentionableOption)
    assert options["f"].name == "f"
    assert options["f"].description == "quuz"
    assert options["f"].is_required

    assert isinstance(options["g"], arc.AttachmentOption)
    assert options["g"].name == "g"
    assert options["g"].description == "among us"
    assert not options["g"].is_required


def test_my_group() -> None:
    group = client.slash_commands["my_group"]
    assert isinstance(group, arc.SlashGroup)

    assert len(group.children) == 2
    assert group.client is client
    assert group.name == "my_group"
    assert group.description == "My group description"
    assert group.default_permissions == hikari.Permissions.ADMINISTRATOR
    assert group.is_dm_enabled is True
    assert group.is_nsfw is False
    assert group.name_localizations == {}
    assert group.description_localizations == {}

    subcmd = group.children["test_subcommand"]

    assert subcmd.client is client
    assert subcmd.parent is group
    assert subcmd.name == "test_subcommand"
    assert subcmd.description == "My subcommand description"
    assert isinstance(subcmd, arc.SlashSubCommand)
    assert isinstance(subcmd.options["a"], arc.IntOption)
    assert isinstance(subcmd.options["b"], arc.StrOption)

    subgroup = group.children["my_subgroup"]

    assert subgroup.client is client
    assert subgroup.parent is group
    assert subgroup.name == "my_subgroup"
    assert subgroup.description == "My subgroup description"
    assert isinstance(subgroup, arc.SlashSubGroup)
    assert len(subgroup.children) == 1

    subsubcmd = subgroup.children["test_subsubcommand"]
    assert subsubcmd.qualified_name == ("my_group", "my_subgroup", "test_subsubcommand")



from __future__ import annotations

import abc
import functools
import inspect
import itertools
import typing as t

import hikari

from .abc import HasErrorHandler
from .command import MessageCommand, SlashCommand, SlashGroup, UserCommand
from .context import AutodeferMode, Context
from .internal.types import BuilderT, ClientT, EventCallbackT, GatewayClientT, RESTClientT, SlashCommandLike

if t.TYPE_CHECKING:
    from .command import CommandBase, SlashSubCommand, SlashSubGroup

__all__ = ("Plugin", "RESTPlugin", "GatewayPlugin")

P = t.ParamSpec("P")
T = t.TypeVar("T")


class Plugin(HasErrorHandler[ClientT], t.Generic[ClientT]):
    """A base class for plugins.

    Parameters
    ----------
    name : builtins.str
        The name of this plugin. This must be unique across all plugins.
    """

    def __init__(
        self, name: str, *, default_enabled_guilds: hikari.UndefinedOr[t.Sequence[hikari.Snowflake]] = hikari.UNDEFINED
    ) -> None:
        super().__init__()
        self._client: ClientT | None = None
        self._name = name
        self._slash_commands: dict[str, SlashCommandLike[ClientT]] = {}
        self._user_commands: dict[str, UserCommand[ClientT]] = {}
        self._message_commands: dict[str, MessageCommand[ClientT]] = {}
        self._default_enabled_guilds = default_enabled_guilds

    @property
    @abc.abstractmethod
    def is_rest(self) -> bool:
        """Whether or not this plugin is a REST plugin."""

    @property
    def name(self) -> str:
        """The name of this plugin."""
        return self._name

    @property
    def client(self) -> ClientT:
        """The client this plugin is included in."""
        if self._client is None:
            raise RuntimeError(
                f"Plugin '{self.name}' was not included in a client, '{type(self).__name__}.client' cannot be accessed until it is included in a client."
            )
        return self._client

    @property
    def default_enabled_guilds(self) -> hikari.UndefinedOr[t.Sequence[hikari.Snowflake]]:
        """The default guilds to enable commands in."""
        return self._default_enabled_guilds

    def _client_include_hook(self, client: ClientT) -> None:
        if client._plugins.get(self.name) is not None:
            raise RuntimeError(f"Plugin '{self.name}' is already included in client.")

        self._client = client
        self._client._plugins[self.name] = self

        for command in itertools.chain(
            self._slash_commands.values(), self._user_commands.values(), self._message_commands.values()
        ):
            command._client_include_hook(client)

    def _client_remove_hook(self) -> None:
        if self._client is None:
            raise RuntimeError(f"Plugin '{self.name}' is not included in a client.")

        for command in itertools.chain(
            self._slash_commands.values(), self._user_commands.values(), self._message_commands.values()
        ):
            self.client._remove_command(command)

        self._client._plugins.pop(self.name)
        self._client = None

    def _add_command(self, command: CommandBase[ClientT, t.Any]) -> None:
        if isinstance(command, (SlashCommand, SlashGroup)):
            self._slash_commands[command.name] = command
        elif isinstance(command, UserCommand):
            self._user_commands[command.name] = command
        elif isinstance(command, MessageCommand):
            self._message_commands[command.name] = command
        else:
            raise TypeError(f"Unknown command type '{type(command).__name__}'.")

    def include(self, command: CommandBase[ClientT, BuilderT]) -> CommandBase[ClientT, BuilderT]:
        """Include a command in this plugin.

        Parameters
        ----------
        command : arc.CommandBase[ClientT, BuilderT]
            The command to include in this plugin.

        Raises
        ------
        RuntimeError
            If the command is already included in this plugin.
        """
        if command.plugin is not None:
            raise ValueError(f"Command '{command.name}' is already registered with plugin '{command.plugin.name}'.")

        command._plugin_include_hook(self)
        return command

    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        try:
            if self.error_handler is not None:
                await self.error_handler(ctx, exc)
            else:
                raise exc
        except Exception as exc:
            await self.client._on_error(ctx, exc)

    def include_slash_group(
        self,
        name: str,
        description: str = "No description provided.",
        *,
        guilds: hikari.UndefinedOr[t.Sequence[hikari.Snowflake]] = hikari.UNDEFINED,
        autodefer: bool | AutodeferMode = True,
        is_dm_enabled: bool = True,
        default_permissions: hikari.UndefinedOr[hikari.Permissions] = hikari.UNDEFINED,
        name_localizations: dict[hikari.Locale, str] | None = None,
        description_localizations: dict[hikari.Locale, str] | None = None,
        is_nsfw: bool = False,
    ) -> SlashGroup[ClientT]:
        """Add a new slash command group to this client.

        Parameters
        ----------
        name : str
            The name of the slash command group.
        description : str
            The description of the slash command group.
        guilds : hikari.UndefinedOr[t.Sequence[hikari.Snowflake]], optional
            The guilds to register the slash command group in, by default hikari.UNDEFINED
        autodefer : bool | AutodeferMode, optional
            If True, all commands in this group will automatically defer if it is taking longer than 2 seconds to respond.
            This can be overridden on a per-subcommand basis.
        is_dm_enabled : bool, optional
            Whether the slash command group is enabled in DMs, by default True
        default_permissions : hikari.UndefinedOr[hikari.Permissions], optional
            The default permissions for the slash command group, by default hikari.UNDEFINED
        name_localizations : dict[hikari.Locale, str], optional
            The name of the slash command group in different locales, by default None
        description_localizations : dict[hikari.Locale, str], optional
            The description of the slash command group in different locales, by default None
        is_nsfw : bool, optional
            Whether the slash command group is only usable in NSFW channels, by default False

        Returns
        -------
        SlashGroup[te.Self]
            The slash command group that was created.

        Usage
        -----
        ```py
        group = client.include_slash_group("Group", "A group of commands.")

        @group.include
        @arc.slash_subcommand(name="Command", description="A command.")
        async def cmd(ctx: arc.Context[arc.GatewayClient]) -> None:
            await ctx.respond("Hello!")
        ```
        """
        children: dict[str, SlashSubCommand[ClientT] | SlashSubGroup[ClientT]] = {}

        group = SlashGroup(
            name=name,
            description=description,
            children=children,
            guilds=guilds,
            autodefer=AutodeferMode(autodefer),
            is_dm_enabled=is_dm_enabled,
            default_permissions=default_permissions,
            name_localizations=name_localizations or {},
            description_localizations=description_localizations or {},
            is_nsfw=is_nsfw,
        )
        group._plugin_include_hook(self)
        return group

    def inject_dependencies(self, func: t.Callable[P, T]) -> t.Callable[P, T]:
        """First order decorator to inject dependencies into the decorated function.

        !!! warning
            This makes functions uncallable if the plugin is not added to a client.

        !!! note
            Command callbacks are automatically injected with dependencies,
            thus this decorator is not needed for them.

        Usage
        -----
        ```py
        class MyDependency:
            def __init__(self, value: str):
                self.value = value

        client = arc.GatewayClient(...)
        client.set_type_dependency(MyDependency, MyDependency("Hello!"))
        client.load_extension("foo")

        # In 'foo':

        plugin = arc.GatewayPlugin("My Plugin")

        @plugin.inject_dependencies
        def my_func(dep: MyDependency = arc.inject()) -> None:
            print(dep.value) # Prints "Hello!"

        @arc.loader
        def load(client: arc.GatewayClient) -> None:
            client.add_plugin(plugin)
        ```

        See Also
        --------
        - [`Client.set_type_dependency`][arc.client.Client.set_type_dependency]
            A method to set dependencies for the client.
        """
        if self._client is None:
            raise RuntimeError(
                f"Cannot inject dependencies into '{func.__name__}' before plugin '{self.name}' is included in a client."
            )

        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def decorator_async(*args: P.args, **kwargs: P.kwargs) -> T:
                return await self.client.injector.call_with_async_di(func, *args, **kwargs)

            return decorator_async  # pyright: ignore reportGeneralTypeIssues
        else:

            @functools.wraps(func)
            def decorator(*args: P.args, **kwargs: P.kwargs) -> T:
                return self.client.injector.call_with_di(func, *args, **kwargs)

            return decorator


class RESTPlugin(Plugin[RESTClientT]):
    """The default implementation of a REST plugin.

    Parameters
    ----------
    name : builtins.str
        The name of this plugin. This must be unique across all plugins.
    """

    @property
    def is_rest(self) -> bool:
        return True


class GatewayPlugin(Plugin[GatewayClientT]):
    """The default implementation of a gateway plugin.

    Parameters
    ----------
    name : builtins.str
        The name of this plugin. This must be unique across all plugins.
    """

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self._listeners: dict[t.Type[hikari.Event], set[EventCallbackT[t.Any]]] = {}

    @property
    def is_rest(self) -> bool:
        return False

    @property
    def listeners(self) -> t.Mapping[t.Type[hikari.Event], t.Collection[EventCallbackT[t.Any]]]:
        return self._listeners

    def _client_include_hook(self, client: GatewayClientT) -> None:
        super()._client_include_hook(client)

        for event, callbacks in self.listeners.items():
            for callback in callbacks:
                client.app.event_manager.subscribe(event, callback)

    def _client_remove_hook(self) -> None:
        if self._client is None:
            raise RuntimeError(f"Plugin '{self.name}' is not included in a client.")

        for event, callbacks in self.listeners.items():
            for callback in callbacks:
                self.client.app.event_manager.unsubscribe(event, callback)

        super()._client_remove_hook()


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

import hikari

from .internal.types import GatewayClientT

if t.TYPE_CHECKING:
    from .context import Context

__all__ = ("ArcEvent", "CommandErrorEvent")


class ArcEvent(hikari.Event):
    """Base class for all Arc events."""

    pass


class CommandErrorEvent(ArcEvent, t.Generic[GatewayClientT]):
    """Event dispatched when a command raises an exception that is not handled by any error handlers."""

    def __init__(self, client: GatewayClientT, context: Context[GatewayClientT], exception: Exception) -> None:
        self._context = context
        self._client = client
        self._exception = exception

    @property
    def app(self) -> hikari.RESTAware:
        """App instance for this application."""
        return self._client.app

    @property
    def context(self) -> Context[GatewayClientT]:
        """The invocation context that raised the exception."""
        return self._context

    @property
    def exception(self) -> Exception:
        """The exception that was raised."""
        return self._exception



from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    import hikari

    from ..client import Client, GatewayClient, RESTClient
    from ..command import OptionParams, SlashCommand, SlashGroup
    from ..context import AutocompleteData, Context


# Generics
AppT = t.TypeVar("AppT", bound="hikari.RESTAware")
ChoiceT = t.TypeVar("ChoiceT", bound="int | float | str")
ClientT = t.TypeVar("ClientT", bound="Client[t.Any]")
GatewayClientT = t.TypeVar("GatewayClientT", bound="GatewayClient")
RESTClientT = t.TypeVar("RESTClientT", bound="RESTClient")
EventT = t.TypeVar("EventT", bound="hikari.Event")
BuilderT = t.TypeVar("BuilderT", bound="hikari.api.SlashCommandBuilder | hikari.api.ContextMenuCommandBuilder")
ParamsT = t.TypeVar("ParamsT", bound="OptionParams[t.Any]")

# Type aliases
EventCallbackT: t.TypeAlias = "t.Callable[[EventT], t.Coroutine[t.Any, t.Any, None]]"
ErrorHandlerCallbackT: t.TypeAlias = "t.Callable[[Context[ClientT], Exception], t.Coroutine[t.Any, t.Any, None]]"
SlashCommandLike: t.TypeAlias = "SlashCommand[ClientT] | SlashGroup[ClientT]"
CommandCallbackT: t.TypeAlias = "t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]]"
MessageContextCallbackT: t.TypeAlias = (
    "t.Callable[t.Concatenate[Context[ClientT], hikari.Message, ...], t.Awaitable[None]]"
)
UserContextCallbackT: t.TypeAlias = "t.Callable[t.Concatenate[Context[ClientT], hikari.User, ...], t.Awaitable[None]]"
AutocompleteCallbackT: t.TypeAlias = "t.Callable[[AutocompleteData[ClientT, ChoiceT]], t.Awaitable[t.Sequence[ChoiceT]]] | t.Callable[[AutocompleteData[ClientT, ChoiceT]], t.Awaitable[t.Sequence[hikari.api.AutocompleteChoiceBuilder]]]"
ResponseBuilderT: t.TypeAlias = (
    "hikari.api.InteractionMessageBuilder | hikari.api.InteractionDeferredBuilder | hikari.api.InteractionModalBuilder"
)



import hikari
import pytest

import arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(bot)


class Dummy:
    def __init__(self, val: int):
        self.val = val


client.set_type_dependency(Dummy, Dummy(0))


@client.inject_dependencies
def dummy_callback(dummy: Dummy = arc.inject()) -> int:
    dummy.val += 1
    return dummy.val


@client.inject_dependencies
async def async_dummy_callback(dummy: Dummy = arc.inject()) -> int:
    dummy.val += 1
    return dummy.val


@pytest.mark.asyncio
async def test_inject() -> None:
    assert dummy_callback() == 1
    assert await async_dummy_callback() == 2
    assert dummy_callback() == 3



from __future__ import annotations

import inspect
import types
import typing as t

import hikari

from ..command.option import (
    AttachmentOption,
    AttachmentParams,
    ChannelOption,
    ChannelParams,
    FloatOption,
    FloatParams,
    IntOption,
    IntParams,
    MentionableOption,
    MentionableParams,
    OptionParams,
    RoleOption,
    RoleParams,
    StrOption,
    StrParams,
    UserOption,
    UserParams,
)

if t.TYPE_CHECKING:
    from ..command.option import CommandOptionBase
    from ..context import Context
    from .types import ClientT


__all__ = ("parse_function_signature",)

TYPE_TO_OPTION_MAPPING: dict[t.Type[t.Any], t.Type[CommandOptionBase[t.Any, t.Any, t.Any]]] = {
    int: IntOption,
    str: StrOption,
    float: FloatOption,
    hikari.User | hikari.Role: MentionableOption,
    t.Union[hikari.User, hikari.Role]: MentionableOption,
    hikari.Attachment: AttachmentOption,
    hikari.User: UserOption,
    hikari.GuildTextChannel: ChannelOption,
    hikari.GuildVoiceChannel: ChannelOption,
    hikari.GuildCategory: ChannelOption,
    hikari.GuildNewsChannel: ChannelOption,
    hikari.GuildPrivateThread: ChannelOption,
    hikari.GuildPublicThread: ChannelOption,
    hikari.GuildForumChannel: ChannelOption,
    hikari.DMChannel: ChannelOption,
    hikari.GroupDMChannel: ChannelOption,
    hikari.GuildStageChannel: ChannelOption,
    hikari.PartialChannel: ChannelOption,
    hikari.TextableChannel: ChannelOption,
    hikari.GuildChannel: ChannelOption,
    hikari.PrivateChannel: ChannelOption,
    hikari.PermissibleGuildChannel: ChannelOption,
    hikari.TextableGuildChannel: ChannelOption,
}

OPT_TO_PARAMS_MAPPING: dict[t.Type[CommandOptionBase[t.Any, t.Any, t.Any]], t.Type[t.Any]] = {
    IntOption: IntParams,
    StrOption: StrParams,
    FloatOption: FloatParams,
    UserOption: UserParams,
    ChannelOption: ChannelParams,
    MentionableOption: MentionableParams,
    RoleOption: RoleParams,
    AttachmentOption: AttachmentParams,
}

# This is totally not cursed in any way
CHANNEL_TYPES_MAPPING: dict[t.Type[hikari.PartialChannel], hikari.ChannelType | set[hikari.ChannelType]] = {
    hikari.GuildTextChannel: hikari.ChannelType.GUILD_TEXT,
    hikari.GuildVoiceChannel: hikari.ChannelType.GUILD_VOICE,
    hikari.GuildCategory: hikari.ChannelType.GUILD_CATEGORY,
    hikari.GuildNewsChannel: hikari.ChannelType.GUILD_NEWS,
    hikari.GuildPrivateThread: hikari.ChannelType.GUILD_PRIVATE_THREAD,
    hikari.GuildPublicThread: hikari.ChannelType.GUILD_PUBLIC_THREAD,
    hikari.GuildForumChannel: hikari.ChannelType.GUILD_FORUM,
    hikari.DMChannel: hikari.ChannelType.DM,
    hikari.GroupDMChannel: hikari.ChannelType.GROUP_DM,
    hikari.GuildStageChannel: hikari.ChannelType.GUILD_STAGE,
    hikari.GuildThreadChannel: {
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
        hikari.ChannelType.GUILD_NEWS_THREAD,
    },
    hikari.PartialChannel: {
        hikari.ChannelType.GUILD_TEXT,
        hikari.ChannelType.GUILD_VOICE,
        hikari.ChannelType.GUILD_CATEGORY,
        hikari.ChannelType.GUILD_NEWS,
        hikari.ChannelType.GUILD_FORUM,
        hikari.ChannelType.GUILD_NEWS_THREAD,
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
        hikari.ChannelType.DM,
        hikari.ChannelType.GROUP_DM,
        hikari.ChannelType.GUILD_STAGE,
    },
    hikari.TextableChannel: {
        hikari.ChannelType.GUILD_TEXT,
        hikari.ChannelType.GUILD_VOICE,
        hikari.ChannelType.GUILD_NEWS,
        hikari.ChannelType.GUILD_NEWS_THREAD,
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
        hikari.ChannelType.DM,
        hikari.ChannelType.GROUP_DM,
        hikari.ChannelType.GUILD_STAGE,
    },
    hikari.GuildChannel: {
        hikari.ChannelType.GUILD_TEXT,
        hikari.ChannelType.GUILD_VOICE,
        hikari.ChannelType.GUILD_CATEGORY,
        hikari.ChannelType.GUILD_NEWS,
        hikari.ChannelType.GUILD_FORUM,
        hikari.ChannelType.GUILD_NEWS_THREAD,
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
        hikari.ChannelType.GUILD_STAGE,
    },
    hikari.PrivateChannel: {hikari.ChannelType.DM, hikari.ChannelType.GROUP_DM},
    hikari.PermissibleGuildChannel: {
        hikari.ChannelType.GUILD_TEXT,
        hikari.ChannelType.GUILD_VOICE,
        hikari.ChannelType.GUILD_CATEGORY,
        hikari.ChannelType.GUILD_FORUM,
        hikari.ChannelType.GUILD_NEWS,
        hikari.ChannelType.GUILD_STAGE,
    },
    hikari.TextableGuildChannel: {
        hikari.ChannelType.GUILD_TEXT,
        hikari.ChannelType.GUILD_VOICE,
        hikari.ChannelType.GUILD_NEWS,
        hikari.ChannelType.GUILD_NEWS_THREAD,
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
        hikari.ChannelType.GUILD_STAGE,
    },
}


def _is_param(meta: t.Any) -> bool:
    return isinstance(meta, OptionParams)


def _is_union(hint: t.Any) -> bool:
    return t.get_origin(hint) is t.Union or t.get_origin(hint) is types.UnionType


def _get_supported_types() -> list[str]:
    """Get a list of supported types.
    Used in error messages.

    Returns
    -------
    list[str]
        The list of supported types
    """
    return [type_.__name__ if type(type_) is type else repr(type_) for type_ in TYPE_TO_OPTION_MAPPING]


def _is_mentionable_union(hint: t.Any) -> bool:
    """Check if a type hint is a union that represents a MentionableOption.

    Parameters
    ----------
    hint : t.Any
        The type hint to check

    Returns
    -------
    bool
        Whether the type hint is a union that represents a Mentionable option
    """
    if not _is_union(hint):
        return False

    return {arg for arg in t.get_args(hint) if arg is not type(None)} == {hikari.User, hikari.Role}


def _channels_to_channel_types(channels: t.Iterable[t.Type[hikari.PartialChannel]]) -> list[hikari.ChannelType]:
    """Turn a list of channels into a list of channel types.

    Parameters
    ----------
    channels : Iterable[hikari.PartialChannel]
        The channels to parse

    Returns
    -------
    list[hikari.ChannelType]
        The list of channel types
    """
    channel_types: set[hikari.ChannelType] = set()

    for channel in channels:
        types = CHANNEL_TYPES_MAPPING.get(channel)

        if types is None:
            raise TypeError(
                f"Unsupported channel type '{channel.__name__}'\nSupported types: {tuple(CHANNEL_TYPES_MAPPING)}"
            )

        if isinstance(types, set):
            channel_types.update(types)
        else:
            channel_types.add(types)

    return list(channel_types)


def _parse_channel_union_type_hint(hint: t.Any) -> list[hikari.ChannelType]:
    """Turn a union of channel types into a list of channel types.

    E.g. `hikari.GuildTextChannel | hikari.DMChannel` becomes `[hikari.ChannelType.GUILD_TEXT, hikari.ChannelType.DM]`

    Parameters
    ----------
    hint : t.Any
        The type hint to parse

    Returns
    -------
    list[hikari.ChannelType]
        The list of channel types

    Raises
    ------
    TypeError
        The type hint was not a union of channel types
    """
    if not _is_union(hint):
        raise TypeError(f"Expected type hint to be 'Union', got '{hint!r}'")

    args = t.get_args(hint)

    if not all((issubclass(arg, hikari.PartialChannel)) or arg is type(None) for arg in args):
        raise TypeError(f"Union expressions are only supported for channels, not '{hint!r}'")

    return _channels_to_channel_types(arg for arg in args if arg is not type(None))


# TODO Detect if param has a default value and also make it optional
def parse_function_signature(  # noqa: C901
    func: t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]],
) -> dict[str, CommandOptionBase[t.Any, t.Any, t.Any]]:
    """Parse a command callback function's signature and return a list of options.

    Parameters
    ----------
    func : Callable[Concatenate[Context, P], None]
        The callback function to parse

    Returns
    -------
    dict[str, Option]
        A mapping of the keyword-argument name to the parsed option

    Raises
    ------
    TypeError
        One of the type hints was not of type 'Annotated'
    TypeError
        The type hint had more than 1 metadata item
    TypeError
        Unsupported option type
    TypeError
        The params type does not match the argument type
    TypeError
        One of the union types was not supported
    """
    hints = t.get_type_hints(func, include_extras=True)
    parameters = inspect.signature(func).parameters

    options: dict[str, CommandOptionBase[t.Any, t.Any, t.Any]] = {}
    # Remove the return type
    hints.pop("return", None)

    for arg_name, hint in hints.items():
        # Ignore non-annotated type hints
        if t.get_origin(hint) is not t.Annotated:
            continue

        if len(hint.__metadata__) != 1:
            continue

        params = hint.__metadata__[0]
        type_ = t.get_args(hint)[0]
        union = None
        is_optional = parameters[arg_name].default is not inspect.Parameter.empty

        # This may be an alluka.Injected or other type of annotation
        if not _is_param(params):
            continue

        # If it's a union, verify all types are supported
        if _is_union(type_):
            union = type_
            union_args = t.get_args(union)

            if not _is_mentionable_union(union) and not all(
                arg is type(None) or arg in TYPE_TO_OPTION_MAPPING for arg in union_args
            ):
                raise TypeError(
                    f"Unsupported option type: '{union!r}'\nSupported option types: {_get_supported_types()}"
                )
            type_ = next((arg for arg in union_args if arg in TYPE_TO_OPTION_MAPPING))
            is_optional = is_optional or type(None) in union_args

        # Verify if it's a supported type
        elif type_ not in TYPE_TO_OPTION_MAPPING:
            raise TypeError(f"Unsupported option type: '{type_!r}'\nSupported option types: {_get_supported_types()}")

        # Get the corresponding option type
        if union is not None and _is_mentionable_union(union):
            opt_type = TYPE_TO_OPTION_MAPPING[union]
        else:
            opt_type = TYPE_TO_OPTION_MAPPING[type_]

        if not isinstance(params, OPT_TO_PARAMS_MAPPING[opt_type]):
            raise TypeError(
                f"Expected params object to be of type {OPT_TO_PARAMS_MAPPING[opt_type].__name__}, got '{type(params).__name__}'"
            )

        # If it's a union of channel types, we need to parse all channel types
        if union is not None and type_ in CHANNEL_TYPES_MAPPING:
            channel_types = _parse_channel_union_type_hint(union)
            options[arg_name] = ChannelOption._from_params(
                name=params.name or arg_name, is_required=not is_optional, params=params, channel_types=channel_types
            )
            continue

        # Parse mentionable unions
        if union is not None and {arg for arg in t.get_args(union) if arg is not type(None)} == {
            hikari.User,
            hikari.Role,
        }:
            options[arg_name] = MentionableOption._from_params(
                name=params.name or arg_name, is_required=not is_optional, params=params
            )
            continue

        # If it's a single channel type, just pass the channel type
        if type_ in CHANNEL_TYPES_MAPPING:
            options[arg_name] = ChannelOption._from_params(
                name=params.name or arg_name,
                is_required=not is_optional,
                params=params,
                channel_types=_channels_to_channel_types([type_]),
            )
            continue

        # Otherwise just build the option
        options[arg_name] = TYPE_TO_OPTION_MAPPING[type_]._from_params(
            name=params.name or arg_name, is_required=not is_optional, params=params
        )

    return options


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t
from functools import cached_property

import attr

from ..internal.types import ChoiceT, ClientT

if t.TYPE_CHECKING:
    import hikari

    from ..command import CommandProto

__all__ = ("AutocompleteData",)


# Don't slot, because we want to use cached_property.
@attr.define(slots=False, kw_only=True)
class AutocompleteData(t.Generic[ClientT, ChoiceT]):
    """The data that is provided to an autocomplete callback."""

    client: ClientT
    """The client that received the autocomplete."""

    command: CommandProto
    """The command that needs to be autocompleted."""

    interaction: hikari.AutocompleteInteraction
    """The interaction that triggered the autocomplete."""

    options: t.Sequence[hikari.AutocompleteInteractionOption]
    """The options that have been provided so far."""

    @cached_property
    def focused_option(self) -> hikari.AutocompleteInteractionOption | None:
        """The option that is currently being focused."""
        return next((o for o in self.options if o.is_focused), None)

    @cached_property
    def focused_value(self) -> ChoiceT | str | None:
        """The value that is currently being focused. This property will return `None` if there is no focused option.

        !!! warning
            This property may not parsed by the client yet, in that case it is returned as a `str`.

        ??? tip
            According to some testing, this option is always either `None` or a string, however
            the API documentation says that it can be the option type as well.
        """
        if self.focused_option is None:
            return None
        return t.cast(ChoiceT | str, self.focused_option.value)

    @property
    def guild_id(self) -> hikari.Snowflake | None:
        """The guild ID of the interaction."""
        return self.interaction.guild_id

    @property
    def channel_id(self) -> hikari.Snowflake:
        """The channel ID of the interaction."""
        return self.interaction.channel_id

    @property
    def member(self) -> hikari.Member | None:
        """The member that triggered the interaction."""
        return self.interaction.member

    @property
    def user(self) -> hikari.User:
        """The user that triggered the interaction."""
        return self.interaction.user

    def get_guild(self) -> hikari.Guild | None:
        """Get the guild that triggered the interaction."""
        return self.interaction.get_guild()

    def get_channel(self) -> hikari.TextableGuildChannel | None:
        """Get the channel that triggered the interaction."""
        return self.interaction.get_channel()


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t
from warnings import warn

from .version import CURRENT_VERSION, Version

__all__ = ("warn_deprecate",)


def warn_deprecate(*, what: str, when: Version, use_instead: t.Optional[str] = None) -> None:
    """Warn about a deprecation.

    Parameters
    ----------
    what : str
        The name of the object that is being deprecated.
    when : Version
        The version in which the object will be removed.
    use_instead : Optional[str], optional
        The object's name that should be used instead, by default None
    """
    if when < CURRENT_VERSION:
        raise DeprecationWarning(
            f"{what!r} is past its deprecation date and should not be used. {f'Use {use_instead!r} instead.' if use_instead else ''}"
        )

    if use_instead:
        warn(
            f"{what!r} is deprecated and will be removed in version {when}. Use {use_instead!r} instead.",
            DeprecationWarning,
            stacklevel=3,
        )
    else:
        warn(f"{what!r} is deprecated and will be removed in version {when}.", DeprecationWarning, stacklevel=3)


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import typing as t

import hikari
import pytest

import arc
from arc.internal.sigparse import parse_function_signature


async def correct_command(
    ctx: arc.Context[t.Any],
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
    c: arc.Option[float | None, arc.FloatParams(description="baz", max=50.0)],
    d: arc.Option[hikari.GuildTextChannel | hikari.GuildNewsChannel | None, arc.ChannelParams(description="qux")],
    e: arc.Option[hikari.GuildChannel | None, arc.ChannelParams(description="quux")],
    f: arc.Option[hikari.Role | hikari.User, arc.MentionableParams(description="quuz")],
    g: arc.Option[hikari.Attachment | None, arc.AttachmentParams(description="among us")],
    i: arc.Option[str, arc.StrParams(description="foo")] = "among us",
) -> None:
    pass


def test_correct_command() -> None:
    options = parse_function_signature(correct_command)
    assert len(options) == 8

    assert isinstance(options["a"], arc.IntOption)
    assert options["a"].name == "a"
    assert options["a"].description == "foo"
    assert options["a"].is_required
    assert options["a"].min == 10
    assert options["a"].max is None

    assert isinstance(options["b"], arc.StrOption)
    assert options["b"].name == "b"
    assert options["b"].description == "bar"
    assert options["b"].is_required
    assert options["b"].min_length == 100
    assert options["b"].max_length is None

    assert isinstance(options["c"], arc.FloatOption)
    assert options["c"].name == "c"
    assert options["c"].description == "baz"
    assert not options["c"].is_required
    assert options["c"].min is None
    assert options["c"].max == 50.0

    assert isinstance(options["d"], arc.ChannelOption)
    assert options["d"].name == "d"
    assert options["d"].description == "qux"
    assert not options["d"].is_required
    assert options["d"].channel_types is not None
    assert set(options["d"].channel_types) == {hikari.ChannelType.GUILD_TEXT, hikari.ChannelType.GUILD_NEWS}

    assert isinstance(options["e"], arc.ChannelOption)
    assert options["e"].name == "e"
    assert options["e"].description == "quux"
    assert not options["e"].is_required
    assert options["e"].channel_types is not None
    assert set(options["e"].channel_types) == {
        hikari.ChannelType.GUILD_TEXT,
        hikari.ChannelType.GUILD_VOICE,
        hikari.ChannelType.GUILD_CATEGORY,
        hikari.ChannelType.GUILD_NEWS,
        hikari.ChannelType.GUILD_FORUM,
        hikari.ChannelType.GUILD_NEWS_THREAD,
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
        hikari.ChannelType.GUILD_STAGE,
    }

    assert isinstance(options["f"], arc.MentionableOption)
    assert options["f"].name == "f"
    assert options["f"].description == "quuz"
    assert options["f"].is_required

    assert isinstance(options["g"], arc.AttachmentOption)
    assert options["g"].name == "g"
    assert options["g"].description == "among us"
    assert not options["g"].is_required

    assert isinstance(options["i"], arc.StrOption)
    assert options["i"].name == "i"
    assert options["i"].description == "foo"
    assert not options["i"].is_required


async def wrong_params_type(
    ctx: arc.Context[t.Any],
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.IntParams(description="bar", max=50)],
) -> None:
    pass


def test_wrong_params_type() -> None:
    with pytest.raises(TypeError):
        parse_function_signature(wrong_params_type)


class WrongType:
    pass


async def wrong_opt_type(ctx: arc.Context[t.Any], a: arc.Option[WrongType, arc.IntParams(description="foo")]) -> None:
    pass


def test_wrong_opt_type() -> None:
    with pytest.raises(TypeError):
        parse_function_signature(wrong_opt_type)


class MyType:
    def __init__(self, val: int) -> None:
        self.val = val


async def di_annotation(
    ctx: arc.Context[arc.GatewayClient],
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Injected[MyType],
    c: arc.Option[str, arc.StrParams(name="b", description="bar", min_length=100)],
) -> None:
    pass


def test_di_annotation() -> None:
    options = parse_function_signature(di_annotation)
    assert len(options) == 2

    assert isinstance(options["a"], arc.IntOption)
    assert options["a"].name == "a"
    assert options["a"].description == "foo"
    assert options["a"].is_required
    assert options["a"].min == 10
    assert options["a"].max is None

    assert isinstance(options["c"], arc.StrOption)
    assert options["c"].name == "b"
    assert options["c"].description == "bar"
    assert options["c"].is_required
    assert options["c"].min_length == 100
    assert options["c"].max_length is None


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import abc
import typing as t

import attr

from ..internal.types import ClientT, ErrorHandlerCallbackT

if t.TYPE_CHECKING:
    from ..context import Context


@attr.define(slots=False)
class HasErrorHandler(abc.ABC, t.Generic[ClientT]):
    _error_handler: ErrorHandlerCallbackT[ClientT] | None = attr.field(default=None)

    """ def __init__(self, error_handler: ErrorHandlerCallbackT[ClientT]) -> None:
        self._error_handler = error_handler """

    @property
    def error_handler(self) -> t.Optional[ErrorHandlerCallbackT[ClientT]]:
        return self._error_handler

    def set_error_handler(self, callback: ErrorHandlerCallbackT[ClientT]) -> ErrorHandlerCallbackT[ClientT]:
        """Decorator to set an error handler for this object. This can be added to commands, groups, or plugins.

        This function will be called when an exception is raised during the invocation of a command.

        Usage
        -----
        ```py
        @client.include
        @arc.slash_command("foo", "Foo command description")
        async def foo(ctx: arc.Context[ClientT]) -> None:
            raise Exception("foo")

        @foo.set_error_handler
        async def foo_error_handler(ctx: arc.Context[ClientT], exc: Exception) -> None:
            await ctx.respond("foo failed")
        ```
        """
        self._error_handler = callback
        return callback

    @abc.abstractmethod
    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        ...
